/*
 * Copyright (c) 1997, 2011, Oracle and/or its affiliates. All rights reserved.
 * Copyright (c) 2015, 2021, Loongson Technology. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 *
 */

#ifndef CPU_LOONGARCH_VM_NATIVEINST_LOONGARCH_HPP
#define CPU_LOONGARCH_VM_NATIVEINST_LOONGARCH_HPP

#include "asm/assembler.hpp"
#include "memory/allocation.hpp"
#include "runtime/icache.hpp"
#include "runtime/os.hpp"
#include "utilities/top.hpp"

// We have interfaces for the following instructions:
// - NativeInstruction
// - - NativeCall
// - - NativeMovConstReg
// - - NativeMovConstRegPatching
// - - NativeMovRegMem
// - - NativeMovRegMemPatching
// - - NativeIllegalOpCode
// - - NativeGeneralJump
// - - NativePushConst
// - - NativeTstRegMem

// The base class for different kinds of native instruction abstractions.
// Provides the primitive operations to manipulate code relative to this.

class NativeInstruction VALUE_OBJ_CLASS_SPEC {
  friend class Relocation;

 public:
  enum loongarch_specific_constants {
    nop_instruction_code        =    0,
    nop_instruction_size        =    4,
    sync_instruction_code       =    0xf
  };

  bool is_nop()                        { guarantee(0, "LA not implemented yet"); return long_at(0) == nop_instruction_code; }
  bool is_sync()                       { return Assembler::high(insn_word(), 17) == Assembler::dbar_op; }
  bool is_dtrace_trap();
  inline bool is_call();
  inline bool is_illegal();
  bool is_jump();
  bool is_safepoint_poll();

  // LoongArch has no instruction to generate a illegal instrucion exception?
  // But mips `break  11` is not illegal instruction for LoongArch.
  static int illegal_instruction();

  bool is_int_branch();
  bool is_float_branch();

  //We use an illegal instruction for marking a method as not_entrant or zombie.
  bool is_sigill_zombie_not_entrant();

 protected:
  address addr_at(int offset) const    { return address(this) + offset; }
  address instruction_address() const       { return addr_at(0); }
  address next_instruction_address() const  { return addr_at(BytesPerInstWord); }
  address prev_instruction_address() const  { return addr_at(-BytesPerInstWord); }

  s_char sbyte_at(int offset) const    { return *(s_char*) addr_at(offset); }
  u_char ubyte_at(int offset) const    { return *(u_char*) addr_at(offset); }

  jint int_at(int offset) const         { return *(jint*) addr_at(offset); }
  juint uint_at(int offset) const       { return *(juint*) addr_at(offset); }

  intptr_t ptr_at(int offset) const    { return *(intptr_t*) addr_at(offset); }

  oop  oop_at (int offset) const       { return *(oop*) addr_at(offset); }
  int  long_at(int offset) const       { return *(jint*)addr_at(offset); }


  void set_char_at(int offset, char c)        { *addr_at(offset) = (u_char)c; wrote(offset); }
  void set_int_at(int offset, jint  i)        { *(jint*)addr_at(offset) = i;  wrote(offset); }
  void set_ptr_at (int offset, intptr_t  ptr) { *(intptr_t*) addr_at(offset) = ptr;  wrote(offset); }
  void set_oop_at (int offset, oop  o)        { *(oop*) addr_at(offset) = o;  wrote(offset); }
  void set_long_at(int offset, long  i);

  int  insn_word() const { return long_at(0); }

  void wrote(int offset);

 public:

  // unit test stuff
  static void test() {}                 // override for testing

  inline friend NativeInstruction* nativeInstruction_at(address address);
};

inline NativeInstruction* nativeInstruction_at(address address) {
  NativeInstruction* inst = (NativeInstruction*)address;
#ifdef ASSERT
  //inst->verify();
#endif
  return inst;
}

// The NativeCall is an abstraction for accessing/manipulating native call imm32/imm64
// instructions (used to manipulate inline caches, primitive & dll calls, etc.).
// LOONGARCH has no call instruction with imm32/imm64. Usually, a call was done like this:
//   short:
//     nop
//     bl offs26
//
//   far:
//     pcaddu18i reg, si20
//     jirl  ra, reg, si18
//
// we just consider the above for instruction as one call instruction
//
inline NativeCall* nativeCall_at(address address);
class NativeCall: public NativeInstruction {
 public:
  enum loongarch_specific_constants {
    instruction_offset    = 0,
    instruction_size      = 2 * BytesPerInstWord,
    return_address_offset = 2 * BytesPerInstWord,
    displacement_offset   = 0
  };

  bool is_short() const;
  bool is_far() const;

  address instruction_address() const { return addr_at(instruction_offset); }

  address next_instruction_address() const {
    return addr_at(return_address_offset);
  }

  address return_address() const {
    return next_instruction_address();
  }

  address destination(address orig_addr = 0) const;
  void set_destination(address dest);
  void set_destination_mt_safe(address dest) { set_destination(dest);}

  void verify_alignment() {}
  void verify();
  void print();

  // Creation
  inline friend NativeCall* nativeCall_at(address address);
  inline friend NativeCall* nativeCall_before(address return_address);

  static bool is_call_at(address instr) {
    return nativeInstruction_at(instr)->is_call();
  }

  static bool is_call_before(address return_address) {
    return is_call_at(return_address - return_address_offset);
  }

  static bool is_call_to(address instr, address target) {
    return nativeInstruction_at(instr)->is_call() &&
           nativeCall_at(instr)->destination() == target;
  }

  // MT-safe patching of a call instruction.
  static void insert(address code_pos, address entry);
  static void replace_mt_safe(address instr_addr, address code_buffer);
};

inline NativeCall* nativeCall_at(address address) {
  NativeCall* call = (NativeCall*)(address - NativeCall::instruction_offset);
#ifdef ASSERT
  call->verify();
#endif
  return call;
}

inline NativeCall* nativeCall_before(address return_address) {
  NativeCall* call = (NativeCall*)(return_address - NativeCall::return_address_offset);
#ifdef ASSERT
  call->verify();
#endif
  return call;
}

class NativeMovConstReg: public NativeInstruction {
 public:
  enum loongarch_specific_constants {
    instruction_offset    =    0,
#ifndef _LP64
    instruction_size          =    2 * BytesPerInstWord,
    next_instruction_offset   =    2 * BytesPerInstWord,
#else
    instruction_size          =    3 * BytesPerInstWord,
    next_instruction_offset   =    3 * BytesPerInstWord,
#endif
  };

  int     insn_word() const                 { return long_at(instruction_offset); }
  address instruction_address() const       { return addr_at(0); }
  address next_instruction_address() const  { return addr_at(next_instruction_offset); }
  intptr_t data() const;
  void    set_data(intptr_t x);

  bool is_li52() const {
    return is_lu12iw_ori_lu32id() ||
           is_lu12iw_lu32id_nop() ||
           is_lu12iw_2nop() ||
           is_lu12iw_ori_nop() ||
           is_addid_2nop();
  }
  bool is_lu12iw_ori_lu32id() const;
  bool is_lu12iw_lu32id_nop() const;
  bool is_lu12iw_2nop() const;
  bool is_lu12iw_ori_nop() const;
  bool is_addid_2nop() const;
  void  verify();
  void  print();

  // unit test stuff
  static void test() {}

  // Creation
  inline friend NativeMovConstReg* nativeMovConstReg_at(address address);
  inline friend NativeMovConstReg* nativeMovConstReg_before(address address);
};

inline NativeMovConstReg* nativeMovConstReg_at(address address) {
  NativeMovConstReg* test = (NativeMovConstReg*)(address - NativeMovConstReg::instruction_offset);
#ifdef ASSERT
  test->verify();
#endif
  return test;
}

inline NativeMovConstReg* nativeMovConstReg_before(address address) {
  NativeMovConstReg* test = (NativeMovConstReg*)(address - NativeMovConstReg::instruction_size - NativeMovConstReg::instruction_offset);
#ifdef ASSERT
  test->verify();
#endif
  return test;
}

class NativeMovConstRegPatching: public NativeMovConstReg {
 private:
    friend NativeMovConstRegPatching* nativeMovConstRegPatching_at(address address) {
    NativeMovConstRegPatching* test = (NativeMovConstRegPatching*)(address - instruction_offset);
    #ifdef ASSERT
      test->verify();
    #endif
    return test;
  }
};

// An interface for accessing/manipulating native moves of the form:
//       lui   AT, split_high(offset)
//       addiu AT, split_low(offset)
//       add   reg, reg, AT
//       lb/lbu/sb/lh/lhu/sh/lw/sw/lwc1/swc1 dest, reg, 0
//       [lw/sw/lwc1/swc1                    dest, reg, 4]
//     or
//       lb/lbu/sb/lh/lhu/sh/lw/sw/lwc1/swc1 dest, reg, offset
//       [lw/sw/lwc1/swc1                    dest, reg, offset+4]
//
// Warning: These routines must be able to handle any instruction sequences
// that are generated as a result of the load/store byte,word,long
// macros.

class NativeMovRegMem: public NativeInstruction {
 public:
  enum loongarch_specific_constants {
    instruction_offset  = 0,
    hiword_offset   = 4,
    ldst_offset     = 12,
    immediate_size  = 4,
    ldst_size       = 16
  };

  address instruction_address() const       { return addr_at(instruction_offset); }
  address next_instruction_address() const  {
    //TODO: LA
    guarantee(0, "LA not implemented yet");
  }

  int   offset() const;

  void  set_offset(int x);

  void  add_offset_in_bytes(int add_offset)     { set_offset ( ( offset() + add_offset ) ); }

  void verify();
  void print ();

  // unit test stuff
  static void test() {}

 private:
  inline friend NativeMovRegMem* nativeMovRegMem_at (address address);
};

inline NativeMovRegMem* nativeMovRegMem_at (address address) {
  NativeMovRegMem* test = (NativeMovRegMem*)(address - NativeMovRegMem::instruction_offset);
#ifdef ASSERT
  test->verify();
#endif
  return test;
}

class NativeMovRegMemPatching: public NativeMovRegMem {
 private:
  friend NativeMovRegMemPatching* nativeMovRegMemPatching_at (address address) {
    NativeMovRegMemPatching* test = (NativeMovRegMemPatching*)(address - instruction_offset);
    #ifdef ASSERT
      test->verify();
    #endif
    return test;
  }
};


// Handles all kinds of jump on Loongson.
//   short:
//     b offs26
//     nop
//
//   far:
//     pcaddu18i reg, si20
//     jirl  r0, reg, si18
//
class NativeGeneralJump: public NativeInstruction {
 public:
  enum loongarch_specific_constants {
    instruction_offset = 0,
    instruction_size   = 2 * BytesPerInstWord
  };

  bool is_short();
  bool is_far();

  address instruction_address() const { return addr_at(instruction_offset); }
  address jump_destination(address orig_addr = 0);
  void  set_jump_destination(address dest);

  // Creation
  inline friend NativeGeneralJump* nativeGeneralJump_at(address address);

  // Insertion of native general jump instruction
  static void insert_unconditional(address code_pos, address entry);
  static void replace_mt_safe(address instr_addr, address code_buffer);
  static void check_verified_entry_alignment(address entry, address verified_entry){}
  static void patch_verified_entry(address entry, address verified_entry, address dest);

  void verify();
};

inline NativeGeneralJump* nativeGeneralJump_at(address address) {
  NativeGeneralJump* jump = (NativeGeneralJump*)(address);
  debug_only(jump->verify();)
  return jump;
}


class NativeIllegalInstruction: public NativeInstruction {
public:
  enum loongarch_specific_constants {
    instruction_code        = 0xbadc0de0, // TODO: LA
                                          // Temporary LoongArch reserved instruction
    instruction_size        = 4,
    instruction_offset      = 0,
    next_instruction_offset = 4
  };

  // Insert illegal opcode as specific address
  static void insert(address code_pos);
};

inline bool NativeInstruction::is_illegal() { return insn_word() == illegal_instruction(); }

inline bool NativeInstruction::is_call() {
  NativeCall *call = (NativeCall*)instruction_address();

  // short
  if (call->is_short()) {
    return true;
  }

  // far
  if (call->is_far()) {
    return true;
  }

  return false;
}

inline bool NativeInstruction::is_jump()
{
  NativeGeneralJump *jump = (NativeGeneralJump*)instruction_address();

  // short
  if (jump->is_short()) {
    return true;
  }

  // far
  if (jump->is_far()) {
    return true;
  }

  return false;
}

#endif // CPU_LOONGARCH_VM_NATIVEINST_LOONGARCH_HPP
