/*
 * Copyright (c) 2003, 2013, Oracle and/or its affiliates. All rights reserved.
 * Copyright (c) 2015, 2021, Loongson Technology. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 *
 */

#include "precompiled.hpp"
#include "asm/macroAssembler.hpp"
#include "asm/macroAssembler.inline.hpp"
#include "interpreter/interpreter.hpp"
#include "nativeInst_loongarch.hpp"
#include "oops/instanceOop.hpp"
#include "oops/method.hpp"
#include "oops/objArrayKlass.hpp"
#include "oops/oop.inline.hpp"
#include "prims/methodHandles.hpp"
#include "runtime/frame.inline.hpp"
#include "runtime/handles.inline.hpp"
#include "runtime/sharedRuntime.hpp"
#include "runtime/stubCodeGenerator.hpp"
#include "runtime/stubRoutines.hpp"
#include "runtime/thread.inline.hpp"
#include "utilities/top.hpp"
#ifdef COMPILER2
#include "opto/runtime.hpp"
#endif

// Declaration and definition of StubGenerator (no .hpp file).
// For a more detailed description of the stub routine structure
// see the comment in stubRoutines.hpp

#define __ _masm->

#define A0 RA0
#define A1 RA1
#define A2 RA2
#define A3 RA3
#define A4 RA4
#define A5 RA5
#define A6 RA6
#define A7 RA7
#define T0 RT0
#define T1 RT1
#define T2 RT2
#define T3 RT3
#define T4 RT4
#define T5 RT5
#define T6 RT6
#define T7 RT7
#define T8 RT8

#define TIMES_OOP (UseCompressedOops ? Address::times_4 : Address::times_8)
//#define a__ ((Assembler*)_masm)->

//#ifdef PRODUCT
//#define BLOCK_COMMENT(str) /* nothing */
//#else
//#define BLOCK_COMMENT(str) __ block_comment(str)
//#endif

//#define BIND(label) bind(label); BLOCK_COMMENT(#label ":")
const int MXCSR_MASK = 0xFFC0;  // Mask out any pending exceptions

// Stub Code definitions

static address handle_unsafe_access() {
  JavaThread* thread = JavaThread::current();
  address pc = thread->saved_exception_pc();
  // pc is the instruction which we must emulate
  // doing a no-op is fine:  return garbage from the load
  // therefore, compute npc
  address npc = (address)((unsigned long)pc + sizeof(unsigned int));

  // request an async exception
  thread->set_pending_unsafe_access_error();

  // return address of next instruction to execute
  return npc;
}

class StubGenerator: public StubCodeGenerator {
 private:

  // ABI mips n64
  // This fig is not MIPS ABI. It is call Java from C ABI.
  // Call stubs are used to call Java from C
  //
  //    [ return_from_Java     ]
  //    [ argument word n-1    ] <--- sp
  //      ...
  //    [ argument word 0      ]
  //      ...
  // -8 [ S6                   ]
  // -7 [ S5                   ]
  // -6 [ S4                   ]
  // -5 [ S3                   ]
  // -4 [ S1                   ]
  // -3 [ TSR(S2)              ]
  // -2 [ LVP(S7)              ]
  // -1 [ BCP(S1)              ]
  //  0 [ saved fp             ] <--- fp_after_call
  //  1 [ return address       ]
  //  2 [ ptr. to call wrapper ] <--- a0 (old sp -->)fp
  //  3 [ result               ] <--- a1
  //  4 [ result_type          ] <--- a2
  //  5 [ method               ] <--- a3
  //  6 [ entry_point          ] <--- a4
  //  7 [ parameters           ] <--- a5
  //  8 [ parameter_size       ] <--- a6
  //  9 [ thread               ] <--- a7

  //
  // _LP64: n64 does not save paras in sp.
  //
  //    [ return_from_Java     ]
  //    [ argument word n-1    ] <--- sp
  //      ...
  //    [ argument word 0      ]
  //      ...
  //-13 [ thread               ]
  //-12 [ result_type          ] <--- a2
  //-11 [ result               ] <--- a1
  //-10 [                      ]
  // -9 [ ptr. to call wrapper ] <--- a0
  // -8 [ S6                   ]
  // -7 [ S5                   ]
  // -6 [ S4                   ]
  // -5 [ S3                   ]
  // -4 [ S1                   ]
  // -3 [ TSR(S2)              ]
  // -2 [ LVP(S7)              ]
  // -1 [ BCP(S1)              ]
  //  0 [ saved fp             ] <--- fp_after_call
  //  1 [ return address       ]
  //  2 [                      ] <--- old sp
  //
  // Find a right place in the call_stub for S8.
  // S8 will point to the starting point of Interpreter::dispatch_table(itos).
  // It should be saved/restored before/after Java calls.
  //
  enum call_stub_layout {
    RA_off             =  1,
    FP_off             =  0,
    BCP_off            = -1,
    LVP_off            = -2,
    TSR_off            = -3,
    S1_off             = -4,
    S3_off             = -5,
    S4_off             = -6,
    S5_off             = -7,
    S6_off             = -8,
    call_wrapper_off   = -9,
    result_off         = -11,
    result_type_off    = -12,
    thread_off         = -13,
    total_off          = thread_off - 1,
    S8_off             = -14,
  };

  address generate_call_stub(address& return_address) {
    assert((int)frame::entry_frame_call_wrapper_offset == (int)call_wrapper_off, "adjust this code");
    StubCodeMark mark(this, "StubRoutines", "call_stub");
    address start = __ pc();

    // same as in generate_catch_exception()!

    // stub code
    // save ra and fp
    __ enter();
    // I think 14 is the max gap between argument and callee saved register
    __ addi_d(SP, SP, total_off * wordSize);
    __ st_d(BCP, FP, BCP_off * wordSize);
    __ st_d(LVP, FP, LVP_off * wordSize);
    __ st_d(TSR, FP, TSR_off * wordSize);
    __ st_d(S1, FP, S1_off * wordSize);
    __ st_d(S3, FP, S3_off * wordSize);
    __ st_d(S4, FP, S4_off * wordSize);
    __ st_d(S5, FP, S5_off * wordSize);
    __ st_d(S6, FP, S6_off * wordSize);
    __ st_d(A0, FP, call_wrapper_off * wordSize);
    __ st_d(A1, FP, result_off * wordSize);
    __ st_d(A2, FP, result_type_off * wordSize);
    __ st_d(A7, FP, thread_off * wordSize);
    __ st_d(S8, FP, S8_off * wordSize);

    __ set64(S8, (long)Interpreter::dispatch_table(itos));

#ifdef OPT_THREAD
    __ move(TREG, A7);
#endif
    //add for compressedoops
    __ reinit_heapbase();

#ifdef ASSERT
    // make sure we have no pending exceptions
    {
      Label L;
      __ ld_d(AT, A7, in_bytes(Thread::pending_exception_offset()));
      __ beq(AT, R0, L);
      /* FIXME: I do not know how to realize stop in mips arch, do it in the future */
      __ stop("StubRoutines::call_stub: entered with pending exception");
      __ bind(L);
    }
#endif

    // pass parameters if any
    // A5: parameter
    // A6: parameter_size
    // T0: parameter_size_tmp(--)
    // T2: offset(++)
    // T3: tmp
    Label parameters_done;
    // judge if the parameter_size equals 0
    __ beq(A6, R0, parameters_done);
    __ slli_d(AT, A6, Interpreter::logStackElementSize);
    __ sub_d(SP, SP, AT);
    __ move(AT, -StackAlignmentInBytes);
    __ andr(SP, SP, AT);
    // Copy Java parameters in reverse order (receiver last)
    // Note that the argument order is inverted in the process
    Label loop;
    __ move(T0, A6);
    __ move(T2, R0);
    __ bind(loop);

    // get parameter
    __ alsl_d(T3, T0, A5, LogBytesPerWord - 1);
    __ ld_d(AT, T3,  -wordSize);
    __ alsl_d(T3, T2, SP, LogBytesPerWord - 1);
    __ st_d(AT, T3, Interpreter::expr_offset_in_bytes(0));
    __ addi_d(T2, T2, 1);
    __ addi_d(T0, T0, -1);
    __ bne(T0, R0, loop);
    // advance to next parameter

    // call Java function
    __ bind(parameters_done);

    // receiver in V0, methodOop in Rmethod

    __ move(Rmethod, A3);
    __ move(Rsender, SP);             //set sender sp
    __ jalr(A4);
    return_address = __ pc();

    Label common_return;
    __ bind(common_return);

    // store result depending on type
    // (everything that is not T_LONG, T_FLOAT or T_DOUBLE is treated as T_INT)
    __ ld_d(T0, FP, result_off * wordSize);   // result --> T0
    Label is_long, is_float, is_double, exit;
    __ ld_d(T2, FP, result_type_off * wordSize);  // result_type --> T2
    __ addi_d(T3, T2, (-1) * T_LONG);
    __ beq(T3, R0, is_long);
    __ addi_d(T3, T2, (-1) * T_FLOAT);
    __ beq(T3, R0, is_float);
    __ addi_d(T3, T2, (-1) * T_DOUBLE);
    __ beq(T3, R0, is_double);

    // handle T_INT case
    __ st_d(V0, T0, 0 * wordSize);
    __ bind(exit);

    // restore
    __ ld_d(BCP, FP, BCP_off * wordSize);
    __ ld_d(LVP, FP, LVP_off * wordSize);
    __ ld_d(S8, FP, S8_off * wordSize);
    __ ld_d(TSR, FP, TSR_off * wordSize);

    __ ld_d(S1, FP, S1_off * wordSize);
    __ ld_d(S3, FP, S3_off * wordSize);
    __ ld_d(S4, FP, S4_off * wordSize);
    __ ld_d(S5, FP, S5_off * wordSize);
    __ ld_d(S6, FP, S6_off * wordSize);

    __ leave();

    // return
    __ jr(RA);

    // handle return types different from T_INT
    __ bind(is_long);
    __ st_d(V0, T0, 0 * wordSize);
    __ b(exit);

    __ bind(is_float);
    __ fst_s(FV0, T0, 0 * wordSize);
    __ b(exit);

    __ bind(is_double);
    __ fst_d(FV0, T0, 0 * wordSize);
    __ b(exit);
    StubRoutines::la::set_call_stub_compiled_return(__ pc());
    __ b(common_return);
    return start;
  }

  // Return point for a Java call if there's an exception thrown in
  // Java code.  The exception is caught and transformed into a
  // pending exception stored in JavaThread that can be tested from
  // within the VM.
  //
  // Note: Usually the parameters are removed by the callee. In case
  // of an exception crossing an activation frame boundary, that is
  // not the case if the callee is compiled code => need to setup the
  // sp.
  //
  // V0: exception oop

  address generate_catch_exception() {
    StubCodeMark mark(this, "StubRoutines", "catch_exception");
    address start = __ pc();

    Register thread = TREG;

    // get thread directly
#ifndef OPT_THREAD
    __ ld_d(thread, FP, thread_off * wordSize);
#endif

#ifdef ASSERT
    // verify that threads correspond
    { Label L;
      __ get_thread(T8);
      __ beq(T8, thread, L);
      __ stop("StubRoutines::catch_exception: threads must correspond");
      __ bind(L);
    }
#endif
    // set pending exception
    __ verify_oop(V0);
    __ st_d(V0, thread, in_bytes(Thread::pending_exception_offset()));
    __ li(AT, (long)__FILE__);
    __ st_d(AT, thread, in_bytes(Thread::exception_file_offset   ()));
    __ li(AT, (long)__LINE__);
    __ st_d(AT, thread, in_bytes(Thread::exception_line_offset   ()));

    // complete return to VM
    assert(StubRoutines::_call_stub_return_address != NULL, "_call_stub_return_address must have been generated before");
    __ jmp(StubRoutines::_call_stub_return_address, relocInfo::none);
    return start;
  }

  // Continuation point for runtime calls returning with a pending
  // exception.  The pending exception check happened in the runtime
  // or native call stub.  The pending exception in Thread is
  // converted into a Java-level exception.
  //
  // Contract with Java-level exception handlers:
  // V0: exception
  // V1: throwing pc
  //
  // NOTE: At entry of this stub, exception-pc must be on stack !!

  address generate_forward_exception() {
    StubCodeMark mark(this, "StubRoutines", "forward exception");
    //Register thread = TREG;
    Register thread = TREG;
    address start = __ pc();

    // Upon entry, the sp points to the return address returning into
    // Java (interpreted or compiled) code; i.e., the return address
    // throwing pc.
    //
    // Arguments pushed before the runtime call are still on the stack
    // but the exception handler will reset the stack pointer ->
    // ignore them.  A potential result in registers can be ignored as
    // well.

#ifndef OPT_THREAD
    __ get_thread(thread);
#endif
#ifdef ASSERT
    // make sure this code is only executed if there is a pending exception
    {
      Label L;
      __ ld_d(AT, thread, in_bytes(Thread::pending_exception_offset()));
      __ bne(AT, R0, L);
      __ stop("StubRoutines::forward exception: no pending exception (1)");
      __ bind(L);
    }
#endif

    // compute exception handler into T4
    __ ld_d(A1, SP, 0);
    __ call_VM_leaf(CAST_FROM_FN_PTR(address, SharedRuntime::exception_handler_for_return_address), thread, A1);
    __ move(T4, V0);
    __ pop(V1);

#ifndef OPT_THREAD
    __ get_thread(thread);
#endif
    __ ld_d(V0, thread, in_bytes(Thread::pending_exception_offset()));
    __ st_d(R0, thread, in_bytes(Thread::pending_exception_offset()));

#ifdef ASSERT
    // make sure exception is set
    {
      Label L;
      __ bne(V0, R0, L);
      __ stop("StubRoutines::forward exception: no pending exception (2)");
      __ bind(L);
    }
#endif

    // continue at exception handler (return address removed)
    // V0: exception
    // T4: exception handler
    // V1: throwing pc
    __ verify_oop(V0);
    __ jr(T4);
    return start;
  }

  // Support for intptr_t get_previous_fp()
  //
  // This routine is used to find the previous frame pointer for the
  // caller (current_frame_guess). This is used as part of debugging
  // ps() is seemingly lost trying to find frames.
  // This code assumes that caller current_frame_guess) has a frame.
  address generate_get_previous_fp() {
    StubCodeMark mark(this, "StubRoutines", "get_previous_fp");
    const Address old_fp       (FP,  0);
    const Address older_fp       (V0,  0);
    address start = __ pc();
    __ enter();
    __ ld_w(V0, old_fp); // callers fp
    __ ld_w(V0, older_fp); // the frame for ps()
    __ leave();
    __ jr(RA);
    return start;
  }

  // The following routine generates a subroutine to throw an
  // asynchronous UnknownError when an unsafe access gets a fault that
  // could not be reasonably prevented by the programmer.  (Example:
  // SIGBUS/OBJERR.)
  address generate_handler_for_unsafe_access() {
    StubCodeMark mark(this, "StubRoutines", "handler_for_unsafe_access");
    address start = __ pc();
    __ push(V0);
    __ pushad_except_v0();                      // push registers
    __ call(CAST_FROM_FN_PTR(address, handle_unsafe_access), relocInfo::runtime_call_type);
    __ popad_except_v0();
    __ move(RA, V0);
    __ pop(V0);
    __ jr(RA);
    return start;
  }

  // Non-destructive plausibility checks for oops
  //
  address generate_verify_oop() {
    StubCodeMark mark(this, "StubRoutines", "verify_oop");
    address start = __ pc();
    __ reinit_heapbase();
    __ verify_oop_subroutine();
    address end = __ pc();
    return start;
  }

  //
  //  Generate overlap test for array copy stubs
  //
  //  Input:
  //     A0    -  array1
  //     A1    -  array2
  //     A2    -  element count
  //

 // use T4 as temp
  void array_overlap_test(address no_overlap_target, int log2_elem_size) {
    int elem_size = 1 << log2_elem_size;
    Address::ScaleFactor sf = Address::times_1;

    switch (log2_elem_size) {
      case 0: sf = Address::times_1; break;
      case 1: sf = Address::times_2; break;
      case 2: sf = Address::times_4; break;
      case 3: sf = Address::times_8; break;
    }
    __ slli_d(AT, A2, sf);
    __ add_d(AT, AT, A0);
    __ addi_d(T4, AT, -elem_size);
    __ sub_d(AT, A1, A0);
    __ bge(R0, AT, no_overlap_target);
    __ sub_d(AT, A1, T4);
    __ blt(R0, AT, no_overlap_target);

    // If A0 = 0xf... and A1 = 0x0..., than goto no_overlap_target
    Label L;
    __ bge(A0, R0, L);
    __ blt(R0, A1, no_overlap_target);
    __ bind(L);
  }

  //
  //  Generate store check for array
  //
  //  Input:
  //     T0    -  starting address
  //     T1    -  element count
  //
  //  The 2 input registers are overwritten
  //


  void array_store_check(Register tmp) {
    assert_different_registers(tmp, AT, T0, T1);
    BarrierSet* bs = Universe::heap()->barrier_set();
    assert(bs->kind() == BarrierSet::CardTableModRef, "Wrong barrier set kind");
    CardTableModRefBS* ct = (CardTableModRefBS*)bs;
    assert(sizeof(*ct->byte_map_base) == sizeof(jbyte), "adjust this code");
    Label l_0, L1;
    if (UseConcMarkSweepGC) {
      __ dbar(0);
    }

    __ set64(tmp, (long)ct->byte_map_base);

    __ alsl_d(AT, T1, T0, TIMES_OOP - 1);
    __ addi_d(T1, AT, -BytesPerHeapOop);

    __ shr(T0, CardTableModRefBS::card_shift);
    __ shr(T1, CardTableModRefBS::card_shift);

    __ sub_d(T1, T1, T0);   // end --> cards count
    __ bind(l_0);

    __ add_d(AT, tmp, T0);
    __ stx_b(R0, AT, T1);
    __ bge(R0, T1, L1);
    __ addi_d(T1, T1, -1);
    __ b(l_0);
    __ bind(L1);
    __ addi_d(T1, T1, -1);
  }

  // Generate code for an array write pre barrier
  //
  //     addr    -  starting address
  //     count   -  element count
  //     tmp     - scratch register
  //
  //     Destroy no registers!
  //
  void  gen_write_ref_array_pre_barrier(Register addr, Register count, bool dest_uninitialized) {
    BarrierSet* bs = Universe::heap()->barrier_set();
    switch (bs->kind()) {
      case BarrierSet::G1SATBCT:
      case BarrierSet::G1SATBCTLogging:
        // With G1, don't generate the call if we statically know that the target in uninitialized
        if (!dest_uninitialized) {
           __ pushad();                      // push registers
           if (count == A0) {
             if (addr == A1) {
               // exactly backwards!!
               //__ xchgptr(c_rarg1, c_rarg0);
               __ move(AT, A0);
               __ move(A0, A1);
               __ move(A1, AT);
             } else {
               __ move(A1, count);
               __ move(A0, addr);
             }
           } else {
             __ move(A0, addr);
             __ move(A1, count);
           }
           __ call_VM_leaf(CAST_FROM_FN_PTR(address, BarrierSet::static_write_ref_array_pre), 2);
           __ popad();
        }
        break;
      case BarrierSet::CardTableModRef:
      case BarrierSet::CardTableExtension:
      case BarrierSet::ModRef:
        break;
      default:
        ShouldNotReachHere();

    }
  }

  //
  // Generate code for an array write post barrier
  //
  //  Input:
  //     start    - register containing starting address of destination array
  //     count    - elements count
  //     scratch  - scratch register
  //
  //  The input registers are overwritten.
  //
  void  gen_write_ref_array_post_barrier(Register start, Register count, Register scratch) {
    assert_different_registers(start, count, scratch, AT);
    BarrierSet* bs = Universe::heap()->barrier_set();
    switch (bs->kind()) {
      case BarrierSet::G1SATBCT:
      case BarrierSet::G1SATBCTLogging:
        {
          __ pushad();             // push registers (overkill)
          if (count == A0) {
            if (start == A1) {
              // exactly backwards!!
              //__ xchgptr(c_rarg1, c_rarg0);
              __ move(AT, A0);
              __ move(A0, A1);
              __ move(A1, AT);
            } else {
              __ move(A1, count);
              __ move(A0, start);
            }
          } else {
            __ move(A0, start);
            __ move(A1, count);
          }
          __ call_VM_leaf(CAST_FROM_FN_PTR(address, BarrierSet::static_write_ref_array_post), 2);
          __ popad();
        }
        break;
      case BarrierSet::CardTableModRef:
      case BarrierSet::CardTableExtension:
        {
          CardTableModRefBS* ct = (CardTableModRefBS*)bs;
          assert(sizeof(*ct->byte_map_base) == sizeof(jbyte), "adjust this code");

          Label L_loop;
          const Register end = count;

          if (UseConcMarkSweepGC) {
            __ dbar(0);
          }

          int64_t disp = (int64_t) ct->byte_map_base;
          __ set64(scratch, disp);

          __ lea(end, Address(start, count, TIMES_OOP, 0));  // end == start+count*oop_size
          __ addi_d(end, end, -BytesPerHeapOop); // end - 1 to make inclusive
          __ shr(start, CardTableModRefBS::card_shift);
          __ shr(end,   CardTableModRefBS::card_shift);
          __ sub_d(end, end, start); // end --> cards count

          __ add_d(start, start, scratch);

          __ bind(L_loop);
          __ stx_b(R0, start, count);
          __ addi_d(count, count, -1);
          __ bge(count, R0, L_loop);
        }
        break;
      default:
        ShouldNotReachHere();
    }
  }

  // Arguments:
  //   aligned - true => Input and output aligned on a HeapWord == 8-byte boundary
  //             ignored
  //   name    - stub name string
  //
  // Inputs:
  //   c_rarg0   - source array address
  //   c_rarg1   - destination array address
  //   c_rarg2   - element count, treated as ssize_t, can be zero
  //
  // If 'from' and/or 'to' are aligned on 4-, 2-, or 1-byte boundaries,
  // we let the hardware handle it.  The one to eight bytes within words,
  // dwords or qwords that span cache line boundaries will still be loaded
  // and stored atomically.
  //
  // Side Effects:
  //   disjoint_byte_copy_entry is set to the no-overlap entry point
  //   used by generate_conjoint_byte_copy().
  //
  address generate_disjoint_byte_copy(bool aligned, const char * name) {
    StubCodeMark mark(this, "StubRoutines", name);
    __ align(CodeEntryAlignment);


    Register tmp1 = T0;
    Register tmp2 = T1;
    Register tmp3 = T3;

    address start = __ pc();
    __ push(tmp1);
    __ push(tmp2);
    __ push(tmp3);
    __ move(tmp1, A0);
    __ move(tmp2, A1);
    __ move(tmp3, A2);


    Label l_1, l_2, l_3, l_4, l_5, l_6, l_7, l_8, l_9, l_10, l_11;
    Label l_debug;

    __ addi_d(AT, tmp3, -9); //why the number is 9 ?
    __ bge(R0, AT, l_9);

    if (!aligned) {
      __ xorr(AT, tmp1, tmp2);
      __ andi(AT, AT, 1);
      __ bne(AT, R0, l_9); // if arrays don't have the same alignment mod 2, do 1 element copy

      __ andi(AT, tmp1, 1);
      __ beq(AT, R0, l_10); //copy 1 enlement if necessary to aligh to 2 bytes

      __ ld_b(AT, tmp1, 0);
      __ addi_d(tmp1, tmp1, 1);
      __ st_b(AT, tmp2, 0);
      __ addi_d(tmp2, tmp2, 1);
      __ addi_d(tmp3, tmp3, -1);
      __ bind(l_10);

      __ xorr(AT, tmp1, tmp2);
      __ andi(AT, AT, 3);
      __ bne(AT, R0, l_1); // if arrays don't have the same alignment mod 4, do 2 elements copy

      // At this point it is guaranteed that both, from and to have the same alignment mod 4.

      // Copy 2 elements if necessary to align to 4 bytes.
      __ andi(AT, tmp1, 3);
      __ beq(AT, R0, l_2);

      __ ld_hu(AT, tmp1, 0);
      __ addi_d(tmp1, tmp1, 2);
      __ st_h(AT, tmp2, 0);
      __ addi_d(tmp2, tmp2, 2);
      __ addi_d(tmp3, tmp3, -2);
      __ bind(l_2);

      // At this point the positions of both, from and to, are at least 4 byte aligned.

      // Copy 4 elements at a time.
      // Align to 8 bytes, but only if both, from and to, have same alignment mod 8.
      __ xorr(AT, tmp1, tmp2);
      __ andi(AT, AT, 7);
      __ bne(AT, R0, l_6); // not same alignment mod 8 -> copy 2, either from or to will be unaligned

      // Copy a 4 elements if necessary to align to 8 bytes.
      __ andi(AT, tmp1, 7);
      __ beq(AT, R0, l_7);

      __ ld_w(AT, tmp1, 0);
      __ addi_d(tmp3, tmp3, -4);
      __ st_w(AT, tmp2, 0);
      { // FasterArrayCopy
        __ addi_d(tmp1, tmp1, 4);
        __ addi_d(tmp2, tmp2, 4);
      }
    }

    __ bind(l_7);

    // Copy 4 elements at a time; either the loads or the stores can
    // be unaligned if aligned == false.

    { // FasterArrayCopy
      __ addi_d(AT, tmp3, -7);
      __ bge(R0, AT, l_6); // copy 4 at a time if less than 4 elements remain

      __ bind(l_8);
      // For Loongson, there is 128-bit memory access. TODO
      __ ld_d(AT, tmp1, 0);
      __ st_d(AT, tmp2, 0);
      __ addi_d(tmp1, tmp1, 8);
      __ addi_d(tmp2, tmp2, 8);
      __ addi_d(tmp3, tmp3, -8);
      __ addi_d(AT, tmp3, -8);
      __ bge(AT, R0, l_8);
    }
    __ bind(l_6);

    // copy 4 bytes at a time
    { // FasterArrayCopy
      __ addi_d(AT, tmp3, -3);
      __ bge(R0, AT, l_1);

      __ bind(l_3);
      __ ld_w(AT, tmp1, 0);
      __ st_w(AT, tmp2, 0);
      __ addi_d(tmp1, tmp1, 4);
      __ addi_d(tmp2, tmp2, 4);
      __ addi_d(tmp3, tmp3, -4);
      __ addi_d(AT, tmp3, -4);
      __ bge(AT, R0, l_3);

    }

    // do 2 bytes copy
    __ bind(l_1);
    {
      __ addi_d(AT, tmp3, -1);
      __ bge(R0, AT, l_9);

      __ bind(l_5);
      __ ld_hu(AT, tmp1, 0);
      __ addi_d(tmp3, tmp3, -2);
      __ st_h(AT, tmp2, 0);
      __ addi_d(tmp1, tmp1, 2);
      __ addi_d(tmp2, tmp2, 2);
      __ addi_d(AT, tmp3, -2);
      __ bge(AT, R0, l_5);
    }

    //do 1 element copy--byte
    __ bind(l_9);
    __ beq(R0, tmp3, l_4);

    {
      __ bind(l_11);
      __ ld_b(AT, tmp1, 0);
      __ addi_d(tmp3, tmp3, -1);
      __ st_b(AT, tmp2, 0);
      __ addi_d(tmp1, tmp1, 1);
      __ addi_d(tmp2, tmp2, 1);
      __ addi_d(AT, tmp3, -1);
      __ bge(AT, R0, l_11);
    }

    __ bind(l_4);
    __ pop(tmp3);
    __ pop(tmp2);
    __ pop(tmp1);

    __ jr(RA);
    return start;
  }

  // Arguments:
  //   aligned - true => Input and output aligned on a HeapWord == 8-byte boundary
  //             ignored
  //   name    - stub name string
  //
  // Inputs:
  //   A0   - source array address
  //   A1   - destination array address
  //   A2   - element count, treated as ssize_t, can be zero
  //
  // If 'from' and/or 'to' are aligned on 4-, 2-, or 1-byte boundaries,
  // we let the hardware handle it.  The one to eight bytes within words,
  // dwords or qwords that span cache line boundaries will still be loaded
  // and stored atomically.
  //
  address generate_conjoint_byte_copy(bool aligned, const char *name) {
    __ align(CodeEntryAlignment);
    StubCodeMark mark(this, "StubRoutines", name);
    address start = __ pc();

    Label l_copy_4_bytes_loop, l_copy_suffix, l_copy_suffix_loop, l_exit;
    Label l_copy_byte, l_from_unaligned, l_unaligned, l_4_bytes_aligned;

    address nooverlap_target = aligned ?
      StubRoutines::arrayof_jbyte_disjoint_arraycopy() :
      StubRoutines::jbyte_disjoint_arraycopy();

    array_overlap_test(nooverlap_target, 0);

    const Register from      = A0;   // source array address
    const Register to        = A1;   // destination array address
    const Register count     = A2;   // elements count
    const Register end_from  = T3;   // source array end address
    const Register end_to    = T0;   // destination array end address
    const Register end_count = T1;   // destination array end address
    __ push(end_from);
    __ push(end_to);
    __ push(end_count);
    __ push(T8);

    // copy from high to low
    __ move(end_count, count);
    __ add_d(end_from, from, end_count);
    __ add_d(end_to, to, end_count);

    // If end_from and end_to has differante alignment, unaligned copy is performed.
    __ andi(AT, end_from, 3);
    __ andi(T8, end_to, 3);
    __ bne(AT, T8, l_copy_byte);

    // First deal with the unaligned data at the top.
    __ bind(l_unaligned);
    __ beq(end_count, R0, l_exit);

    __ andi(AT, end_from, 3);
    __ bne(AT, R0, l_from_unaligned);

    __ andi(AT, end_to, 3);
    __ beq(AT, R0, l_4_bytes_aligned);

    __ bind(l_from_unaligned);
    __ ld_b(AT, end_from, -1);
    __ st_b(AT, end_to, -1);
    __ addi_d(end_from, end_from, -1);
    __ addi_d(end_to, end_to, -1);
    __ addi_d(end_count, end_count, -1);
    __ b(l_unaligned);

    // now end_to, end_from point to 4-byte aligned high-ends
    //     end_count contains byte count that is not copied.
    // copy 4 bytes at a time
    __ bind(l_4_bytes_aligned);

    __ move(T8, end_count);
    __ addi_d(AT, end_count, -3);
    __ bge(R0, AT, l_copy_suffix);

    //__ andi(T8, T8, 3);
    __ lea(end_from, Address(end_from, -4));
    __ lea(end_to, Address(end_to, -4));

    __ srli_d(end_count, end_count, 2);
    __ align(16);
    __ bind(l_copy_4_bytes_loop); //l_copy_4_bytes
    __ ld_w(AT, end_from, 0);
    __ st_w(AT, end_to, 0);
    __ addi_d(end_from, end_from, -4);
    __ addi_d(end_to, end_to, -4);
    __ addi_d(end_count, end_count, -1);
    __ bne(end_count, R0, l_copy_4_bytes_loop);

    __ b(l_copy_suffix);
    // copy dwords aligned or not with repeat move
    // l_copy_suffix
    // copy suffix (0-3 bytes)
    __ bind(l_copy_suffix);
    __ andi(T8, T8, 3);
    __ beq(T8, R0, l_exit);
    __ addi_d(end_from, end_from, 3);
    __ addi_d(end_to, end_to, 3);
    __ bind(l_copy_suffix_loop);
    __ ld_b(AT, end_from, 0);
    __ st_b(AT, end_to, 0);
    __ addi_d(end_from, end_from, -1);
    __ addi_d(end_to, end_to, -1);
    __ addi_d(T8, T8, -1);
    __ bne(T8, R0, l_copy_suffix_loop);

    __ bind(l_copy_byte);
    __ beq(end_count, R0, l_exit);
    __ ld_b(AT, end_from, -1);
    __ st_b(AT, end_to, -1);
    __ addi_d(end_from, end_from, -1);
    __ addi_d(end_to, end_to, -1);
    __ addi_d(end_count, end_count, -1);
    __ b(l_copy_byte);

    __ bind(l_exit);
    __ pop(T8);
    __ pop(end_count);
    __ pop(end_to);
    __ pop(end_from);
    __ jr(RA);
    return start;
  }

  // Generate stub for disjoint short copy.  If "aligned" is true, the
  // "from" and "to" addresses are assumed to be heapword aligned.
  //
  // Arguments for generated stub:
  //      from:  A0
  //      to:    A1
  //  elm.count: A2 treated as signed
  //  one element: 2 bytes
  //
  // Strategy for aligned==true:
  //
  //  If length <= 9:
  //     1. copy 1 elements at a time (l_5)
  //
  //  If length > 9:
  //     1. copy 4 elements at a time until less than 4 elements are left (l_7)
  //     2. copy 2 elements at a time until less than 2 elements are left (l_6)
  //     3. copy last element if one was left in step 2. (l_1)
  //
  //
  // Strategy for aligned==false:
  //
  //  If length <= 9: same as aligned==true case
  //
  //  If length > 9:
  //     1. continue with step 7. if the alignment of from and to mod 4
  //        is different.
  //     2. align from and to to 4 bytes by copying 1 element if necessary
  //     3. at l_2 from and to are 4 byte aligned; continue with
  //        6. if they cannot be aligned to 8 bytes because they have
  //        got different alignment mod 8.
  //     4. at this point we know that both, from and to, have the same
  //        alignment mod 8, now copy one element if necessary to get
  //        8 byte alignment of from and to.
  //     5. copy 4 elements at a time until less than 4 elements are
  //        left; depending on step 3. all load/stores are aligned.
  //     6. copy 2 elements at a time until less than 2 elements are
  //        left. (l_6)
  //     7. copy 1 element at a time. (l_5)
  //     8. copy last element if one was left in step 6. (l_1)

  address generate_disjoint_short_copy(bool aligned, const char * name) {
    StubCodeMark mark(this, "StubRoutines", name);
    __ align(CodeEntryAlignment);

    Register tmp1 = T0;
    Register tmp2 = T1;
    Register tmp3 = T3;
    Register tmp4 = T8;
    Register tmp5 = T4;
    Register tmp6 = T2;

    address start = __ pc();
    __ push(tmp1);
    __ push(tmp2);
    __ push(tmp3);
    __ move(tmp1, A0);
    __ move(tmp2, A1);
    __ move(tmp3, A2);

    Label l_1, l_2, l_3, l_4, l_5, l_6, l_7, l_8, l_9, l_10, l_11, l_12, l_13, l_14;
    Label l_debug;
    // don't try anything fancy if arrays don't have many elements
    __ addi_d(AT, tmp3, -23);
    __ bge(R0, AT, l_14);
    // move push here
    __ push(tmp4);
    __ push(tmp5);
    __ push(tmp6);

    if (!aligned) {
      __ xorr(AT, A0, A1);
      __ andi(AT, AT, 1);
      __ bne(AT, R0, l_debug); // if arrays don't have the same alignment mod 2, can this happen?

      __ xorr(AT, A0, A1);
      __ andi(AT, AT, 3);
      __ bne(AT, R0, l_1); // if arrays don't have the same alignment mod 4, do 1 element copy

      // At this point it is guaranteed that both, from and to have the same alignment mod 4.

      // Copy 1 element if necessary to align to 4 bytes.
      __ andi(AT, A0, 3);
      __ beq(AT, R0, l_2);

      __ ld_hu(AT, tmp1, 0);
      __ addi_d(tmp1, tmp1, 2);
      __ st_h(AT, tmp2, 0);
      __ addi_d(tmp2, tmp2, 2);
      __ addi_d(tmp3, tmp3, -1);
      __ bind(l_2);

      // At this point the positions of both, from and to, are at least 4 byte aligned.

      // Copy 4 elements at a time.
      // Align to 8 bytes, but only if both, from and to, have same alignment mod 8.
      __ xorr(AT, tmp1, tmp2);
      __ andi(AT, AT, 7);
      __ bne(AT, R0, l_6); // not same alignment mod 8 -> copy 2, either from or to will be unaligned

      // Copy a 2-element word if necessary to align to 8 bytes.
      __ andi(AT, tmp1, 7);
      __ beq(AT, R0, l_7);

      __ ld_w(AT, tmp1, 0);
      __ addi_d(tmp3, tmp3, -2);
      __ st_w(AT, tmp2, 0);
      __ addi_d(tmp1, tmp1, 4);
      __ addi_d(tmp2, tmp2, 4);
    }// end of if (!aligned)

    __ bind(l_7);
    // At this time the position of both, from and to, are at least 8 byte aligned.
    // Copy 8 elemnets at a time.
    // Align to 16 bytes, but only if both from and to have same alignment mod 8.
    __ xorr(AT, tmp1, tmp2);
    __ andi(AT, AT, 15);
    __ bne(AT, R0, l_9);

    // Copy 4-element word if necessary to align to 16 bytes,
    __ andi(AT, tmp1, 15);
    __ beq(AT, R0, l_10);

    __ ld_d(AT, tmp1, 0);
    __ addi_d(tmp3, tmp3, -4);
    __ st_d(AT, tmp2, 0);
    __ addi_d(tmp1, tmp1, 8);
    __ addi_d(tmp2, tmp2, 8);

    __ bind(l_10);

    // Copy 8 elements at a time; either the loads or the stores can
    // be unalligned if aligned == false

    { // FasterArrayCopy
      __ bind(l_11);
      __ ld_d(AT, tmp1, 0);
      __ ld_d(tmp4, tmp1, 8);
      __ ld_d(tmp5, tmp1, 16);
      __ ld_d(tmp6, tmp1, 24);
      __ addi_d(tmp1, tmp1, 32);
      __ st_d(AT, tmp2, 0);
      __ st_d(tmp4, tmp2, 8);
      __ st_d(tmp5, tmp2, 16);
      __ st_d(tmp6, tmp2, 24);
      __ addi_d(tmp2, tmp2, 32);
      __ addi_d(tmp3, tmp3, -16);
      __ addi_d(AT, tmp3, -16);
      __ bge(AT, R0, l_11);
    }
    __ bind(l_9);

    // Copy 4 elements at a time; either the loads or the stores can
    // be unaligned if aligned == false.
    { // FasterArrayCopy
      __ addi_d(AT, tmp3, -15);// loop unrolling 4 times, so if the elements should not be less than 16
      __ bge(R0, AT, l_4); // copy 2 at a time if less than 16 elements remain

      __ bind(l_8);
      __ ld_d(AT, tmp1, 0);
      __ ld_d(tmp4, tmp1, 8);
      __ ld_d(tmp5, tmp1, 16);
      __ ld_d(tmp6, tmp1, 24);
      __ st_d(AT, tmp2, 0);
      __ st_d(tmp4, tmp2, 8);
      __ st_d(tmp5, tmp2,16);
      __ addi_d(tmp1, tmp1, 32);
      __ addi_d(tmp2, tmp2, 32);
      __ addi_d(tmp3, tmp3, -16);
      __ addi_d(AT, tmp3, -16);
      __ st_d(tmp6, tmp2, -8);
      __ bge(AT, R0, l_8);
    }
    __ bind(l_6);

    // copy 2 element at a time
    { // FasterArrayCopy
      __ addi_d(AT, tmp3, -7);
      __ bge(R0, AT, l_4);

      __ bind(l_3);
      __ ld_w(AT, tmp1, 0);
      __ ld_w(tmp4, tmp1, 4);
      __ ld_w(tmp5, tmp1, 8);
      __ ld_w(tmp6, tmp1, 12);
      __ st_w(AT, tmp2, 0);
      __ st_w(tmp4, tmp2, 4);
      __ st_w(tmp5, tmp2, 8);
      __ addi_d(tmp1, tmp1, 16);
      __ addi_d(tmp2, tmp2, 16);
      __ addi_d(tmp3, tmp3, -8);
      __ addi_d(AT, tmp3, -8);
      __ st_w(tmp6, tmp2, -4);
      __ bge(AT, R0, l_3);
    }

    __ bind(l_1);
    // do single element copy (8 bit), can this happen?
    { // FasterArrayCopy
      __ addi_d(AT, tmp3, -3);
      __ bge(R0, AT, l_4);

      __ bind(l_5);
      __ ld_hu(AT, tmp1, 0);
      __ ld_hu(tmp4, tmp1, 2);
      __ ld_hu(tmp5, tmp1, 4);
      __ ld_hu(tmp6, tmp1, 6);
      __ st_h(AT, tmp2, 0);
      __ st_h(tmp4, tmp2, 2);
      __ st_h(tmp5, tmp2, 4);
      __ addi_d(tmp1, tmp1, 8);
      __ addi_d(tmp2, tmp2, 8);
      __ addi_d(tmp3, tmp3, -4);
      __ addi_d(AT, tmp3, -4);
      __ st_h(tmp6, tmp2, -2);
      __ bge(AT, R0, l_5);
    }
    // single element
    __ bind(l_4);

    __ pop(tmp6);
    __ pop(tmp5);
    __ pop(tmp4);

    __ bind(l_14);
    { // FasterArrayCopy
      __ beq(R0, tmp3, l_13);

      __ bind(l_12);
      __ ld_hu(AT, tmp1, 0);
      __ st_h(AT, tmp2, 0);
      __ addi_d(tmp1, tmp1, 2);
      __ addi_d(tmp2, tmp2, 2);
      __ addi_d(tmp3, tmp3, -1);
      __ addi_d(AT, tmp3, -1);
      __ bge(AT, R0, l_12);
    }

    __ bind(l_13);
    __ pop(tmp3);
    __ pop(tmp2);
    __ pop(tmp1);

    __ jr(RA);

    __ bind(l_debug);
    __ stop("generate_disjoint_short_copy should not reach here");
    return start;
  }

  // Arguments:
  //   aligned - true => Input and output aligned on a HeapWord == 8-byte boundary
  //             ignored
  //   name    - stub name string
  //
  // Inputs:
  //   c_rarg0   - source array address
  //   c_rarg1   - destination array address
  //   c_rarg2   - element count, treated as ssize_t, can be zero
  //
  // If 'from' and/or 'to' are aligned on 4- or 2-byte boundaries, we
  // let the hardware handle it.  The two or four words within dwords
  // or qwords that span cache line boundaries will still be loaded
  // and stored atomically.
  //
  address generate_conjoint_short_copy(bool aligned, const char *name) {
    StubCodeMark mark(this, "StubRoutines", name);
    __ align(CodeEntryAlignment);
    address start = __ pc();

    Label l_exit, l_copy_short, l_from_unaligned, l_unaligned, l_4_bytes_aligned;

    address nooverlap_target = aligned ?
            StubRoutines::arrayof_jshort_disjoint_arraycopy() :
            StubRoutines::jshort_disjoint_arraycopy();

    array_overlap_test(nooverlap_target, 1);

    const Register from      = A0;   // source array address
    const Register to        = A1;   // destination array address
    const Register count     = A2;   // elements count
    const Register end_from  = T3;   // source array end address
    const Register end_to    = T0;   // destination array end address
    const Register end_count = T1;   // destination array end address
    __ push(end_from);
    __ push(end_to);
    __ push(end_count);
    __ push(T8);

    // copy from high to low
    __ move(end_count, count);
    __ slli_d(AT, end_count, Address::times_2);
    __ add_d(end_from, from, AT);
    __ add_d(end_to, to, AT);

    // If end_from and end_to has differante alignment, unaligned copy is performed.
    __ andi(AT, end_from, 3);
    __ andi(T8, end_to, 3);
    __ bne(AT, T8, l_copy_short);

    // First deal with the unaligned data at the top.
    __ bind(l_unaligned);
    __ beq(end_count, R0, l_exit);

    __ andi(AT, end_from, 3);
    __ bne(AT, R0, l_from_unaligned);

    __ andi(AT, end_to, 3);
    __ beq(AT, R0, l_4_bytes_aligned);

    // Copy 1 element if necessary to align to 4 bytes.
    __ bind(l_from_unaligned);
    __ ld_hu(AT, end_from, -2);
    __ st_h(AT, end_to, -2);
    __ addi_d(end_from, end_from, -2);
    __ addi_d(end_to, end_to, -2);
    __ addi_d(end_count, end_count, -1);
    __ b(l_unaligned);

    // now end_to, end_from point to 4-byte aligned high-ends
    //     end_count contains byte count that is not copied.
    // copy 4 bytes at a time
    __ bind(l_4_bytes_aligned);

    __ addi_d(AT, end_count, -1);
    __ bge(R0, AT, l_copy_short);

    __ ld_w(AT, end_from, -4);
    __ st_w(AT, end_to, -4);
    __ addi_d(end_from, end_from, -4);
    __ addi_d(end_to, end_to, -4);
    __ addi_d(end_count, end_count, -2);
    __ b(l_4_bytes_aligned);

    // copy 1 element at a time
    __ bind(l_copy_short);
    __ beq(end_count, R0, l_exit);
    __ ld_hu(AT, end_from, -2);
    __ st_h(AT, end_to, -2);
    __ addi_d(end_from, end_from, -2);
    __ addi_d(end_to, end_to, -2);
    __ addi_d(end_count, end_count, -1);
    __ b(l_copy_short);

    __ bind(l_exit);
    __ pop(T8);
    __ pop(end_count);
    __ pop(end_to);
    __ pop(end_from);
    __ jr(RA);
    return start;
  }

  // Arguments:
  //   aligned - true => Input and output aligned on a HeapWord == 8-byte boundary
  //             ignored
  //   is_oop  - true => oop array, so generate store check code
  //   name    - stub name string
  //
  // Inputs:
  //   c_rarg0   - source array address
  //   c_rarg1   - destination array address
  //   c_rarg2   - element count, treated as ssize_t, can be zero
  //
  // If 'from' and/or 'to' are aligned on 4-byte boundaries, we let
  // the hardware handle it.  The two dwords within qwords that span
  // cache line boundaries will still be loaded and stored atomicly.
  //
  // Side Effects:
  //   disjoint_int_copy_entry is set to the no-overlap entry point
  //   used by generate_conjoint_int_oop_copy().
  //
  address generate_disjoint_int_oop_copy(bool aligned, bool is_oop, const char *name, bool dest_uninitialized = false) {
    Label l_3, l_4, l_5, l_6, l_7;
    StubCodeMark mark(this, "StubRoutines", name);

    __ align(CodeEntryAlignment);
    address start = __ pc();
    __ push(T3);
    __ push(T0);
    __ push(T1);
    __ push(T8);
    __ push(T4);
    __ move(T1, A2);
    __ move(T3, A0);
    __ move(T0, A1);

    if (is_oop) {
      gen_write_ref_array_pre_barrier(A1, A2, dest_uninitialized);
    }

    if(!aligned) {
      __ xorr(AT, T3, T0);
      __ andi(AT, AT, 7);
      __ bne(AT, R0, l_5); // not same alignment mod 8 -> copy 1 element each time

      __ andi(AT, T3, 7);
      __ beq(AT, R0, l_6); //copy 2 elements each time

      __ ld_w(AT, T3, 0);
      __ addi_d(T1, T1, -1);
      __ st_w(AT, T0, 0);
      __ addi_d(T3, T3, 4);
      __ addi_d(T0, T0, 4);
    }

    {
      __ bind(l_6);
      __ addi_d(AT, T1, -1);
      __ bge(R0, AT, l_5);

      __ bind(l_7);
      __ ld_d(AT, T3, 0);
      __ st_d(AT, T0, 0);
      __ addi_d(T3, T3, 8);
      __ addi_d(T0, T0, 8);
      __ addi_d(T1, T1, -2);
      __ addi_d(AT, T1, -2);
      __ bge(AT, R0, l_7);
    }

    __ bind(l_5);
    __ beq(T1, R0, l_4);

    __ align(16);
    __ bind(l_3);
    __ ld_w(AT, T3, 0);
    __ st_w(AT, T0, 0);
    __ addi_d(T3, T3, 4);
    __ addi_d(T0, T0, 4);
    __ addi_d(T1, T1, -1);
    __ bne(T1, R0, l_3);

    // exit
    __ bind(l_4);
    if (is_oop) {
      gen_write_ref_array_post_barrier(A1, A2, T1);
    }
    __ pop(T4);
    __ pop(T8);
    __ pop(T1);
    __ pop(T0);
    __ pop(T3);
    __ jr(RA);
    return start;
  }

  // Arguments:
  //   aligned - true => Input and output aligned on a HeapWord == 8-byte boundary
  //             ignored
  //   is_oop  - true => oop array, so generate store check code
  //   name    - stub name string
  //
  // Inputs:
  //   c_rarg0   - source array address
  //   c_rarg1   - destination array address
  //   c_rarg2   - element count, treated as ssize_t, can be zero
  //
  // If 'from' and/or 'to' are aligned on 4-byte boundaries, we let
  // the hardware handle it.  The two dwords within qwords that span
  // cache line boundaries will still be loaded and stored atomicly.
  //
  address generate_conjoint_int_oop_copy(bool aligned, bool is_oop, const char *name, bool dest_uninitialized = false) {
    Label l_2, l_4;
    StubCodeMark mark(this, "StubRoutines", name);
    __ align(CodeEntryAlignment);
    address start = __ pc();
    address nooverlap_target;

    if (is_oop) {
      nooverlap_target = aligned ?
              StubRoutines::arrayof_oop_disjoint_arraycopy() :
              StubRoutines::oop_disjoint_arraycopy();
    } else {
      nooverlap_target = aligned ?
              StubRoutines::arrayof_jint_disjoint_arraycopy() :
              StubRoutines::jint_disjoint_arraycopy();
    }

    array_overlap_test(nooverlap_target, 2);

    if (is_oop) {
      gen_write_ref_array_pre_barrier(A1, A2, dest_uninitialized);
    }
    __ push(T3);
    __ push(T0);
    __ push(T1);
    __ push(T8); // Useless?
    __ push(T4); // Ditto

    __ move(T1, A2);
    __ move(T3, A0);
    __ move(T0, A1);

    // T3: source array address
    // T0: destination array address
    // T1: element count

    __ alsl_d(AT, T1, T3, Address::times_4 - 1);
    __ addi_d(T3, AT, -4);
    __ alsl_d(AT, T1, T0, Address::times_4 - 1);
    __ addi_d(T0, AT, -4);

    __ beq(T1, R0, l_4);

    __ align(16);
    __ bind(l_2);
    __ ld_w(AT, T3, 0);
    __ st_w(AT, T0, 0);
    __ addi_d(T3, T3, -4);
    __ addi_d(T0, T0, -4);
    __ addi_d(T1, T1, -1);
    __ bne(T1, R0, l_2);

    __ bind(l_4);
    if (is_oop) {
      gen_write_ref_array_post_barrier(A1, A2, T1);
    }
    __ pop(T4); // Useless?
    __ pop(T8); // Ditto
    __ pop(T1);
    __ pop(T0);
    __ pop(T3);
    __ jr(RA);
    return start;
  }

  // Arguments:
  //   aligned - true => Input and output aligned on a HeapWord == 8-byte boundary
  //             ignored
  //   is_oop  - true => oop array, so generate store check code
  //   name    - stub name string
  //
  // Inputs:
  //   c_rarg0   - source array address
  //   c_rarg1   - destination array address
  //   c_rarg2   - element count, treated as ssize_t, can be zero
  //
  // If 'from' and/or 'to' are aligned on 4-byte boundaries, we let
  // the hardware handle it.  The two dwords within qwords that span
  // cache line boundaries will still be loaded and stored atomicly.
  //
  // Side Effects:
  //   disjoint_int_copy_entry is set to the no-overlap entry point
  //   used by generate_conjoint_int_oop_copy().
  //
  address generate_disjoint_long_oop_copy(bool aligned, bool is_oop, const char *name, bool dest_uninitialized = false) {
    Label l_3, l_4;
    StubCodeMark mark(this, "StubRoutines", name);
    __ align(CodeEntryAlignment);
    address start = __ pc();

    if (is_oop) {
      gen_write_ref_array_pre_barrier(A1, A2, dest_uninitialized);
    }
    __ push(T3);
    __ push(T0);
    __ push(T1);
    __ push(T8);
    __ push(T4);

    __ move(T1, A2);
    __ move(T3, A0);
    __ move(T0, A1);

    // T3: source array address
    // T0: destination array address
    // T1: element count

    __ beq(T1, R0, l_4);

    __ align(16);
    __ bind(l_3);
    __ ld_d(AT, T3, 0);
    __ st_d(AT, T0, 0);
    __ addi_d(T3, T3, 8);
    __ addi_d(T0, T0, 8);
    __ addi_d(T1, T1, -1);
    __ bne(T1, R0, l_3);

    // exit
    __ bind(l_4);
    if (is_oop) {
      gen_write_ref_array_post_barrier(A1, A2, T1);
    }
    __ pop(T4);
    __ pop(T8);
    __ pop(T1);
    __ pop(T0);
    __ pop(T3);
    __ jr(RA);
    return start;
  }

  // Arguments:
  //   aligned - true => Input and output aligned on a HeapWord == 8-byte boundary
  //             ignored
  //   is_oop  - true => oop array, so generate store check code
  //   name    - stub name string
  //
  // Inputs:
  //   c_rarg0   - source array address
  //   c_rarg1   - destination array address
  //   c_rarg2   - element count, treated as ssize_t, can be zero
  //
  // If 'from' and/or 'to' are aligned on 4-byte boundaries, we let
  // the hardware handle it.  The two dwords within qwords that span
  // cache line boundaries will still be loaded and stored atomicly.
  //
  address generate_conjoint_long_oop_copy(bool aligned, bool is_oop, const char *name, bool dest_uninitialized = false) {
    Label l_2, l_4;
    StubCodeMark mark(this, "StubRoutines", name);
    __ align(CodeEntryAlignment);
    address start = __ pc();
    address nooverlap_target;

    if (is_oop) {
      nooverlap_target = aligned ?
              StubRoutines::arrayof_oop_disjoint_arraycopy() :
              StubRoutines::oop_disjoint_arraycopy();
    } else {
      nooverlap_target = aligned ?
              StubRoutines::arrayof_jlong_disjoint_arraycopy() :
              StubRoutines::jlong_disjoint_arraycopy();
    }

    array_overlap_test(nooverlap_target, 3);

    if (is_oop) {
      gen_write_ref_array_pre_barrier(A1, A2, dest_uninitialized);
    }
    __ push(T3);
    __ push(T0);
    __ push(T1);
    __ push(T8);
    __ push(T4);

    __ move(T1, A2);
    __ move(T3, A0);
    __ move(T0, A1);

    __ alsl_d(AT, T1, T3, Address::times_8 - 1);
    __ addi_d(T3, AT, -8);
    __ alsl_d(AT, T1, T0, Address::times_8 - 1);
    __ addi_d(T0, AT, -8);

    __ beq(T1, R0, l_4);

    __ align(16);
    __ bind(l_2);
    __ ld_d(AT, T3, 0);
    __ st_d(AT, T0, 0);
    __ addi_d(T3, T3, -8);
    __ addi_d(T0, T0, -8);
    __ addi_d(T1, T1, -1);
    __ bne(T1, R0, l_2);

    // exit
    __ bind(l_4);
    if (is_oop) {
      gen_write_ref_array_post_barrier(A1, A2, T1);
    }
    __ pop(T4);
    __ pop(T8);
    __ pop(T1);
    __ pop(T0);
    __ pop(T3);
    __ jr(RA);
    return start;
  }

  //FIXME
  address generate_disjoint_long_copy(bool aligned, const char *name) {
    Label l_1, l_2;
    StubCodeMark mark(this, "StubRoutines", name);
    __ align(CodeEntryAlignment);
    address start = __ pc();
    __ move(T1, A2);
    __ move(T3, A0);
    __ move(T0, A1);
    __ push(T3);
    __ push(T0);
    __ push(T1);
    __ b(l_2);
    __ align(16);
    __ bind(l_1);
    __ ld_d(AT, T3, 0);
    __ st_d (AT, T0, 0);
    __ addi_d(T3, T3, 8);
    __ addi_d(T0, T0, 8);
    __ bind(l_2);
    __ addi_d(T1, T1, -1);
    __ bge(T1, R0, l_1);
    __ pop(T1);
    __ pop(T0);
    __ pop(T3);
    __ jr(RA);
    return start;
  }


  address generate_conjoint_long_copy(bool aligned, const char *name) {
    Label l_1, l_2;
    StubCodeMark mark(this, "StubRoutines", name);
    __ align(CodeEntryAlignment);
    address start = __ pc();
    address nooverlap_target = aligned ?
      StubRoutines::arrayof_jlong_disjoint_arraycopy() :
      StubRoutines::jlong_disjoint_arraycopy();
    array_overlap_test(nooverlap_target, 3);
    __ push(T3);
    __ push(T0);
    __ push(T1);

    __ move(T1, A2);
    __ move(T3, A0);
    __ move(T0, A1);
    __ alsl_d(AT, T1, T3, Address::times_8 - 1);
    __ addi_d(T3, AT, -8);
    __ alsl_d(AT, T1, T0, Address::times_8 - 1);
    __ addi_d(T0, AT, -8);

    __ b(l_2);
    __ align(16);
    __ bind(l_1);
    __ ld_d(AT, T3, 0);
    __ st_d (AT, T0, 0);
    __ addi_d(T3, T3, -8);
    __ addi_d(T0, T0,-8);
    __ bind(l_2);
    __ addi_d(T1, T1, -1);
    __ bge(T1, R0, l_1);
    __ pop(T1);
    __ pop(T0);
    __ pop(T3);
    __ jr(RA);
    return start;
  }

  void generate_arraycopy_stubs() {
    if (UseCompressedOops) {
      StubRoutines::_oop_disjoint_arraycopy          = generate_disjoint_int_oop_copy(false, true,
                                                                                      "oop_disjoint_arraycopy");
      StubRoutines::_oop_arraycopy                   = generate_conjoint_int_oop_copy(false, true,
                                                                                      "oop_arraycopy");
      StubRoutines::_oop_disjoint_arraycopy_uninit   = generate_disjoint_int_oop_copy(false, true,
                                                                                      "oop_disjoint_arraycopy_uninit", true);
      StubRoutines::_oop_arraycopy_uninit            = generate_conjoint_int_oop_copy(false, true,
                                                                                      "oop_arraycopy_uninit", true);
    } else {
      StubRoutines::_oop_disjoint_arraycopy          = generate_disjoint_long_oop_copy(false, true,
                                                                                       "oop_disjoint_arraycopy");
      StubRoutines::_oop_arraycopy                   = generate_conjoint_long_oop_copy(false, true,
                                                                                       "oop_arraycopy");
      StubRoutines::_oop_disjoint_arraycopy_uninit   = generate_disjoint_long_oop_copy(false, true,
                                                                                       "oop_disjoint_arraycopy_uninit", true);
      StubRoutines::_oop_arraycopy_uninit            = generate_conjoint_long_oop_copy(false, true,
                                                                                       "oop_arraycopy_uninit", true);
    }

    StubRoutines::_jbyte_disjoint_arraycopy          = generate_disjoint_byte_copy(false, "jbyte_disjoint_arraycopy");
    StubRoutines::_jshort_disjoint_arraycopy         = generate_disjoint_short_copy(false, "jshort_disjoint_arraycopy");
    StubRoutines::_jint_disjoint_arraycopy           = generate_disjoint_int_oop_copy(false, false, "jint_disjoint_arraycopy");
    StubRoutines::_jlong_disjoint_arraycopy          = generate_disjoint_long_copy(false, "jlong_disjoint_arraycopy");

    StubRoutines::_jbyte_arraycopy  = generate_conjoint_byte_copy(false, "jbyte_arraycopy");
    StubRoutines::_jshort_arraycopy = generate_conjoint_short_copy(false, "jshort_arraycopy");
    StubRoutines::_jint_arraycopy   = generate_conjoint_int_oop_copy(false, false, "jint_arraycopy");
    StubRoutines::_jlong_arraycopy  = generate_conjoint_long_copy(false, "jlong_arraycopy");

    // We don't generate specialized code for HeapWord-aligned source
    // arrays, so just use the code we've already generated
    StubRoutines::_arrayof_jbyte_disjoint_arraycopy  = StubRoutines::_jbyte_disjoint_arraycopy;
    StubRoutines::_arrayof_jbyte_arraycopy           = StubRoutines::_jbyte_arraycopy;

    StubRoutines::_arrayof_jshort_disjoint_arraycopy = StubRoutines::_jshort_disjoint_arraycopy;
    StubRoutines::_arrayof_jshort_arraycopy          = StubRoutines::_jshort_arraycopy;

    StubRoutines::_arrayof_jint_disjoint_arraycopy   = StubRoutines::_jint_disjoint_arraycopy;
    StubRoutines::_arrayof_jint_arraycopy            = StubRoutines::_jint_arraycopy;

    StubRoutines::_arrayof_jlong_disjoint_arraycopy  = StubRoutines::_jlong_disjoint_arraycopy;
    StubRoutines::_arrayof_jlong_arraycopy           = StubRoutines::_jlong_arraycopy;

    StubRoutines::_arrayof_oop_disjoint_arraycopy    = StubRoutines::_oop_disjoint_arraycopy;
    StubRoutines::_arrayof_oop_arraycopy             = StubRoutines::_oop_arraycopy;

    StubRoutines::_arrayof_oop_disjoint_arraycopy_uninit    = StubRoutines::_oop_disjoint_arraycopy_uninit;
    StubRoutines::_arrayof_oop_arraycopy_uninit             = StubRoutines::_oop_arraycopy_uninit;
  }

  // add a function to implement SafeFetch32 and SafeFetchN
  void generate_safefetch(const char* name, int size, address* entry,
                          address* fault_pc, address* continuation_pc) {
    // safefetch signatures:
    //   int      SafeFetch32(int*      adr, int      errValue);
    //   intptr_t SafeFetchN (intptr_t* adr, intptr_t errValue);
    //
    // arguments:
    //   A0 = adr
    //   A1 = errValue
    //
    // result:
    //   PPC_RET  = *adr or errValue
    StubCodeMark mark(this, "StubRoutines", name);

    // Entry point, pc or function descriptor.
    *entry = __ pc();

    // Load *adr into A1, may fault.
    *fault_pc = __ pc();
    switch (size) {
      case 4:
        // int32_t
        __ ld_w(A1, A0, 0);
        break;
      case 8:
        // int64_t
        __ ld_d(A1, A0, 0);
        break;
      default:
        ShouldNotReachHere();
    }

    // return errValue or *adr
    *continuation_pc = __ pc();
    __ add_d(V0, A1, R0);
    __ jr(RA);
  }


#undef __
#define __ masm->

  // Continuation point for throwing of implicit exceptions that are
  // not handled in the current activation. Fabricates an exception
  // oop and initiates normal exception dispatching in this
  // frame. Since we need to preserve callee-saved values (currently
  // only for C2, but done for C1 as well) we need a callee-saved oop
  // map and therefore have to make these stubs into RuntimeStubs
  // rather than BufferBlobs.  If the compiler needs all registers to
  // be preserved between the fault point and the exception handler
  // then it must assume responsibility for that in
  // AbstractCompiler::continuation_for_implicit_null_exception or
  // continuation_for_implicit_division_by_zero_exception. All other
  // implicit exceptions (e.g., NullPointerException or
  // AbstractMethodError on entry) are either at call sites or
  // otherwise assume that stack unwinding will be initiated, so
  // caller saved registers were assumed volatile in the compiler.
  address generate_throw_exception(const char* name,
                                   address runtime_entry,
                                   bool restore_saved_exception_pc) {
    // Information about frame layout at time of blocking runtime call.
    // Note that we only have to preserve callee-saved registers since
    // the compilers are responsible for supplying a continuation point
    // if they expect all registers to be preserved.
    enum layout {
      thread_off,    // last_java_sp
      S7_off,        // callee saved register      sp + 1
      S6_off,        // callee saved register      sp + 2
      S5_off,        // callee saved register      sp + 3
      S4_off,        // callee saved register      sp + 4
      S3_off,        // callee saved register      sp + 5
      S2_off,        // callee saved register      sp + 6
      S1_off,        // callee saved register      sp + 7
      S0_off,        // callee saved register      sp + 8
      FP_off,
      ret_address,
      framesize
    };

    int insts_size = 2048;
    int locs_size  = 32;

    //  CodeBuffer* code     = new CodeBuffer(insts_size, locs_size, 0, 0, 0, false,
    //  NULL, NULL, NULL, false, NULL, name, false);
    CodeBuffer code (name , insts_size, locs_size);
    OopMapSet* oop_maps  = new OopMapSet();
    MacroAssembler* masm = new MacroAssembler(&code);

    address start = __ pc();

    // This is an inlined and slightly modified version of call_VM
    // which has the ability to fetch the return PC out of
    // thread-local storage and also sets up last_Java_sp slightly
    // differently than the real call_VM
#ifndef OPT_THREAD
    Register java_thread = TREG;
    __ get_thread(java_thread);
#else
    Register java_thread = TREG;
#endif
    if (restore_saved_exception_pc) {
      __ ld_d(RA, java_thread, in_bytes(JavaThread::saved_exception_pc_offset()));
    }
    __ enter(); // required for proper stackwalking of RuntimeStub frame

    __ addi_d(SP, SP, (-1) * (framesize-2) * wordSize); // prolog
    __ st_d(S0, SP, S0_off * wordSize);
    __ st_d(S1, SP, S1_off * wordSize);
    __ st_d(S2, SP, S2_off * wordSize);
    __ st_d(S3, SP, S3_off * wordSize);
    __ st_d(S4, SP, S4_off * wordSize);
    __ st_d(S5, SP, S5_off * wordSize);
    __ st_d(S6, SP, S6_off * wordSize);
    __ st_d(S7, SP, S7_off * wordSize);

    int frame_complete = __ pc() - start;
    // push java thread (becomes first argument of C function)
    __ st_d(java_thread, SP, thread_off * wordSize);
    if (java_thread != A0)
      __ move(A0, java_thread);

    // Set up last_Java_sp and last_Java_fp
    Label before_call;
    address the_pc = __ pc();
    __ bind(before_call);
    __ set_last_Java_frame(java_thread, SP, FP, before_call);
    // Align stack
    __ set64(AT, -(StackAlignmentInBytes));
    __ andr(SP, SP, AT);

    // Call runtime
    // TODO: confirm reloc
    __ call(runtime_entry, relocInfo::runtime_call_type);
    // Generate oop map
    OopMap* map =  new OopMap(framesize, 0);
    oop_maps->add_gc_map(the_pc - start,  map);

    // restore the thread (cannot use the pushed argument since arguments
    // may be overwritten by C code generated by an optimizing compiler);
    // however can use the register value directly if it is callee saved.
#ifndef OPT_THREAD
    __ get_thread(java_thread);
#endif

    __ ld_d(SP, java_thread, in_bytes(JavaThread::last_Java_sp_offset()));
    __ reset_last_Java_frame(java_thread, true);

    // Restore callee save registers.  This must be done after resetting the Java frame
    __ ld_d(S0, SP, S0_off * wordSize);
    __ ld_d(S1, SP, S1_off * wordSize);
    __ ld_d(S2, SP, S2_off * wordSize);
    __ ld_d(S3, SP, S3_off * wordSize);
    __ ld_d(S4, SP, S4_off * wordSize);
    __ ld_d(S5, SP, S5_off * wordSize);
    __ ld_d(S6, SP, S6_off * wordSize);
    __ ld_d(S7, SP, S7_off * wordSize);

    // discard arguments
    __ move(SP, FP); // epilog
    __ pop(FP);
    // check for pending exceptions
#ifdef ASSERT
    Label L;
    __ ld_w(AT, java_thread, in_bytes(Thread::pending_exception_offset()));
    __ bne(AT, R0, L);
    __ should_not_reach_here();
    __ bind(L);
#endif //ASSERT
    __ jmp(StubRoutines::forward_exception_entry(), relocInfo::runtime_call_type);

    RuntimeStub* stub = RuntimeStub::new_runtime_stub(name,
                                                      &code,
                                                      frame_complete,
                                                      framesize,
                                                      oop_maps, false);
    return stub->entry_point();
  }

  class MontgomeryMultiplyGenerator : public MacroAssembler {

    Register Pa_base, Pb_base, Pn_base, Pm_base, inv, Rlen, Rlen2, Ra, Rb, Rm,
      Rn, Iam, Ibn, Rhi_ab, Rlo_ab, Rhi_mn, Rlo_mn, t0, t1, t2, Ri, Rj;

    bool _squaring;

  public:
    MontgomeryMultiplyGenerator (Assembler *as, bool squaring)
      : MacroAssembler(as->code()), _squaring(squaring) {

      // Register allocation

      Register reg = A0;
      Pa_base = reg;      // Argument registers:
      if (squaring)
        Pb_base = Pa_base;
      else
        Pb_base = ++reg;
      Pn_base = ++reg;
      Rlen = ++reg;
      inv = ++reg;
      Rlen2 = inv;        // Reuse inv
      Pm_base = ++reg;

                          // Working registers:
      Ra = ++reg;         // The current digit of a, b, n, and m.
      Rb = ++reg;
      Rm = ++reg;
      Rn = ++reg;

      Iam = ++reg;        // Index to the current/next digit of a, b, n, and m.
      Ibn = ++reg;

      t0 = ++reg;         // Three registers which form a
      t1 = ++reg;         // triple-precision accumuator.
      t2 = ++reg;

      Ri = ++reg;         // Inner and outer loop indexes.
      Rj = ++reg;

      if (squaring) {
        Rhi_ab = ++reg;   // Product registers: low and high parts
        reg = S0;
        Rlo_ab = ++reg;   // of a*b and m*n.
      } else {
        reg = S0;
        Rhi_ab = reg;     // Product registers: low and high parts
        Rlo_ab = ++reg;   // of a*b and m*n.
      }

      Rhi_mn = ++reg;
      Rlo_mn = ++reg;
    }

  private:
    void enter() {
      addi_d(SP, SP, -6 * wordSize);
      st_d(FP, SP, 0 * wordSize);
      move(FP, SP);
    }

    void leave() {
      addi_d(T0, FP, 6 * wordSize);
      ld_d(FP, FP, 0 * wordSize);
      move(SP, T0);
    }

    void save_regs() {
      if (!_squaring)
        st_d(Rhi_ab, FP, 5 * wordSize);
      st_d(Rlo_ab, FP, 4 * wordSize);
      st_d(Rhi_mn, FP, 3 * wordSize);
      st_d(Rlo_mn, FP, 2 * wordSize);
      st_d(Pm_base, FP, 1 * wordSize);
    }

    void restore_regs() {
      if (!_squaring)
        ld_d(Rhi_ab, FP, 5 * wordSize);
      ld_d(Rlo_ab, FP, 4 * wordSize);
      ld_d(Rhi_mn, FP, 3 * wordSize);
      ld_d(Rlo_mn, FP, 2 * wordSize);
      ld_d(Pm_base, FP, 1 * wordSize);
    }

    template <typename T>
    void unroll_2(Register count, T block, Register tmp) {
      Label loop, end, odd;
      andi(tmp, count, 1);
      bnez(tmp, odd);
      beqz(count, end);
      align(16);
      bind(loop);
      (this->*block)();
      bind(odd);
      (this->*block)();
      addi_w(count, count, -2);
      blt(R0, count, loop);
      bind(end);
    }

    template <typename T>
    void unroll_2(Register count, T block, Register d, Register s, Register tmp) {
      Label loop, end, odd;
      andi(tmp, count, 1);
      bnez(tmp, odd);
      beqz(count, end);
      align(16);
      bind(loop);
      (this->*block)(d, s, tmp);
      bind(odd);
      (this->*block)(d, s, tmp);
      addi_w(count, count, -2);
      blt(R0, count, loop);
      bind(end);
    }

    void acc(Register Rhi, Register Rlo,
             Register t0, Register t1, Register t2, Register t, Register c) {
      add_d(t0, t0, Rlo);
      OR(t, t1, Rhi);
      sltu(c, t0, Rlo);
      add_d(t1, t1, Rhi);
      add_d(t1, t1, c);
      sltu(c, t1, t);
      add_d(t2, t2, c);
    }

    void pre1(Register i) {
      block_comment("pre1");
      // Iam = 0;
      // Ibn = i;

      slli_w(Ibn, i, LogBytesPerWord);

      // Ra = Pa_base[Iam];
      // Rb = Pb_base[Ibn];
      // Rm = Pm_base[Iam];
      // Rn = Pn_base[Ibn];

      ld_d(Ra, Pa_base, 0);
      ldx_d(Rb, Pb_base, Ibn);
      ld_d(Rm, Pm_base, 0);
      ldx_d(Rn, Pn_base, Ibn);

      move(Iam, R0);

      // Zero the m*n result.
      move(Rhi_mn, R0);
      move(Rlo_mn, R0);
    }

    // The core multiply-accumulate step of a Montgomery
    // multiplication.  The idea is to schedule operations as a
    // pipeline so that instructions with long latencies (loads and
    // multiplies) have time to complete before their results are
    // used.  This most benefits in-order implementations of the
    // architecture but out-of-order ones also benefit.
    void step() {
      block_comment("step");
      // MACC(Ra, Rb, t0, t1, t2);
      // Ra = Pa_base[++Iam];
      // Rb = Pb_base[--Ibn];
      addi_d(Iam, Iam, wordSize);
      addi_d(Ibn, Ibn, -wordSize);
      mul_d(Rlo_ab, Ra, Rb);
      mulh_du(Rhi_ab, Ra, Rb);
      acc(Rhi_mn, Rlo_mn, t0, t1, t2, Ra, Rb); // The pending m*n from the
                                               // previous iteration.
      ldx_d(Ra, Pa_base, Iam);
      ldx_d(Rb, Pb_base, Ibn);

      // MACC(Rm, Rn, t0, t1, t2);
      // Rm = Pm_base[Iam];
      // Rn = Pn_base[Ibn];
      mul_d(Rlo_mn, Rm, Rn);
      mulh_du(Rhi_mn, Rm, Rn);
      acc(Rhi_ab, Rlo_ab, t0, t1, t2, Rm, Rn);
      ldx_d(Rm, Pm_base, Iam);
      ldx_d(Rn, Pn_base, Ibn);
    }

    void post1() {
      block_comment("post1");

      // MACC(Ra, Rb, t0, t1, t2);
      mul_d(Rlo_ab, Ra, Rb);
      mulh_du(Rhi_ab, Ra, Rb);
      acc(Rhi_mn, Rlo_mn, t0, t1, t2, Ra, Rb);  // The pending m*n
      acc(Rhi_ab, Rlo_ab, t0, t1, t2, Ra, Rb);

      // Pm_base[Iam] = Rm = t0 * inv;
      mul_d(Rm, t0, inv);
      stx_d(Rm, Pm_base, Iam);

      // MACC(Rm, Rn, t0, t1, t2);
      // t0 = t1; t1 = t2; t2 = 0;
      mulh_du(Rhi_mn, Rm, Rn);

#ifndef PRODUCT
      // assert(m[i] * n[0] + t0 == 0, "broken Montgomery multiply");
      {
        mul_d(Rlo_mn, Rm, Rn);
        add_d(Rlo_mn, t0, Rlo_mn);
        Label ok;
        beqz(Rlo_mn, ok); {
          stop("broken Montgomery multiply");
        } bind(ok);
      }
#endif

      // We have very carefully set things up so that
      // m[i]*n[0] + t0 == 0 (mod b), so we don't have to calculate
      // the lower half of Rm * Rn because we know the result already:
      // it must be -t0.  t0 + (-t0) must generate a carry iff
      // t0 != 0.  So, rather than do a mul and an adds we just set
      // the carry flag iff t0 is nonzero.
      //
      // mul_d(Rlo_mn, Rm, Rn);
      // add_d(t0, t0, Rlo_mn);
      OR(Ra, t1, Rhi_mn);
      sltu(Rb, R0, t0);
      add_d(t0, t1, Rhi_mn);
      add_d(t0, t0, Rb);
      sltu(Rb, t0, Ra);
      add_d(t1, t2, Rb);
      move(t2, R0);
    }

    void pre2(Register i, Register len) {
      block_comment("pre2");

      // Rj == i-len
      sub_w(Rj, i, len);

      // Iam = i - len;
      // Ibn = len;
      slli_w(Iam, Rj, LogBytesPerWord);
      slli_w(Ibn, len, LogBytesPerWord);

      // Ra = Pa_base[++Iam];
      // Rb = Pb_base[--Ibn];
      // Rm = Pm_base[++Iam];
      // Rn = Pn_base[--Ibn];
      addi_d(Iam, Iam, wordSize);
      addi_d(Ibn, Ibn, -wordSize);

      ldx_d(Ra, Pa_base, Iam);
      ldx_d(Rb, Pb_base, Ibn);
      ldx_d(Rm, Pm_base, Iam);
      ldx_d(Rn, Pn_base, Ibn);

      move(Rhi_mn, R0);
      move(Rlo_mn, R0);
    }

    void post2(Register i, Register len) {
      block_comment("post2");

      sub_w(Rj, i, len);
      slli_w(Iam, Rj, LogBytesPerWord);

      add_d(t0, t0, Rlo_mn); // The pending m*n, low part

      // As soon as we know the least significant digit of our result,
      // store it.
      // Pm_base[i-len] = t0;
      stx_d(t0, Pm_base, Iam);

      // t0 = t1; t1 = t2; t2 = 0;
      OR(Ra, t1, Rhi_mn);
      sltu(Rb, t0, Rlo_mn);
      add_d(t0, t1, Rhi_mn); // The pending m*n, high part
      add_d(t0, t0, Rb);
      sltu(Rb, t0, Ra);
      add_d(t1, t2, Rb);
      move(t2, R0);
    }

    // A carry in t0 after Montgomery multiplication means that we
    // should subtract multiples of n from our result in m.  We'll
    // keep doing that until there is no carry.
    void normalize(Register len) {
      block_comment("normalize");
      // while (t0)
      //   t0 = sub(Pm_base, Pn_base, t0, len);
      Label loop, post, again;
      Register cnt = t1, i = t2, b = Ra, t = Rb; // Re-use registers; we're done with them now
      beqz(t0, post); {
        bind(again); {
          move(i, R0);
          move(b, R0);
          slli_w(cnt, len, LogBytesPerWord);
          align(16);
          bind(loop); {
            ldx_d(Rm, Pm_base, i);
            ldx_d(Rn, Pn_base, i);
            sltu(t, Rm, b);
            sub_d(Rm, Rm, b);
            sltu(b, Rm, Rn);
            sub_d(Rm, Rm, Rn);
            OR(b, b, t);
            stx_d(Rm, Pm_base, i);
            addi_w(i, i, BytesPerWord);
          } blt(i, cnt, loop);
          sub_d(t0, t0, b);
        } bnez(t0, again);
      } bind(post);
    }

    // Move memory at s to d, reversing words.
    //    Increments d to end of copied memory
    //    Destroys tmp1, tmp2, tmp3
    //    Preserves len
    //    Leaves s pointing to the address which was in d at start
    void reverse(Register d, Register s, Register len, Register tmp1, Register tmp2) {
      assert(tmp1 < S0 && tmp2 < S0, "register corruption");

      alsl_d(s, len, s, LogBytesPerWord - 1);
      move(tmp1, len);
      unroll_2(tmp1, &MontgomeryMultiplyGenerator::reverse1, d, s, tmp2);
      slli_w(s, len, LogBytesPerWord);
      sub_d(s, d, s);
    }

    // where
    void reverse1(Register d, Register s, Register tmp) {
      ld_d(tmp, s, -wordSize);
      addi_d(s, s, -wordSize);
      addi_d(d, d, wordSize);
      rotri_d(tmp, tmp, 32);
      st_d(tmp, d, -wordSize);
    }

  public:
    /**
     * Fast Montgomery multiplication.  The derivation of the
     * algorithm is in A Cryptographic Library for the Motorola
     * DSP56000, Dusse and Kaliski, Proc. EUROCRYPT 90, pp. 230-237.
     *
     * Arguments:
     *
     * Inputs for multiplication:
     *   A0   - int array elements a
     *   A1   - int array elements b
     *   A2   - int array elements n (the modulus)
     *   A3   - int length
     *   A4   - int inv
     *   A5   - int array elements m (the result)
     *
     * Inputs for squaring:
     *   A0   - int array elements a
     *   A1   - int array elements n (the modulus)
     *   A2   - int length
     *   A3   - int inv
     *   A4   - int array elements m (the result)
     *
     */
    address generate_multiply() {
      Label argh, nothing;
      bind(argh);
      stop("MontgomeryMultiply total_allocation must be <= 8192");

      align(CodeEntryAlignment);
      address entry = pc();

      beqz(Rlen, nothing);

      enter();

      // Make room.
      sltui(Ra, Rlen, 513);
      beqz(Ra, argh);
      slli_w(Ra, Rlen, exact_log2(4 * sizeof (jint)));
      sub_d(Ra, SP, Ra);

      srli_w(Rlen, Rlen, 1); // length in longwords = len/2

      {
        // Copy input args, reversing as we go.  We use Ra as a
        // temporary variable.
        reverse(Ra, Pa_base, Rlen, t0, t1);
        if (!_squaring)
          reverse(Ra, Pb_base, Rlen, t0, t1);
        reverse(Ra, Pn_base, Rlen, t0, t1);
      }

      // Push all call-saved registers and also Pm_base which we'll need
      // at the end.
      save_regs();

#ifndef PRODUCT
      // assert(inv * n[0] == -1UL, "broken inverse in Montgomery multiply");
      {
        ld_d(Rn, Pn_base, 0);
        li(t0, -1);
        mul_d(Rlo_mn, Rn, inv);
        Label ok;
        beq(Rlo_mn, t0, ok); {
          stop("broken inverse in Montgomery multiply");
        } bind(ok);
      }
#endif

      move(Pm_base, Ra);

      move(t0, R0);
      move(t1, R0);
      move(t2, R0);

      block_comment("for (int i = 0; i < len; i++) {");
      move(Ri, R0); {
        Label loop, end;
        bge(Ri, Rlen, end);

        bind(loop);
        pre1(Ri);

        block_comment("  for (j = i; j; j--) {"); {
          move(Rj, Ri);
          unroll_2(Rj, &MontgomeryMultiplyGenerator::step, Rlo_ab);
        } block_comment("  } // j");

        post1();
        addi_w(Ri, Ri, 1);
        blt(Ri, Rlen, loop);
        bind(end);
        block_comment("} // i");
      }

      block_comment("for (int i = len; i < 2*len; i++) {");
      move(Ri, Rlen);
      slli_w(Rlen2, Rlen, 1); {
        Label loop, end;
        bge(Ri, Rlen2, end);

        bind(loop);
        pre2(Ri, Rlen);

        block_comment("  for (j = len*2-i-1; j; j--) {"); {
          sub_w(Rj, Rlen2, Ri);
          addi_w(Rj, Rj, -1);
          unroll_2(Rj, &MontgomeryMultiplyGenerator::step, Rlo_ab);
        } block_comment("  } // j");

        post2(Ri, Rlen);
        addi_w(Ri, Ri, 1);
        blt(Ri, Rlen2, loop);
        bind(end);
      }
      block_comment("} // i");

      normalize(Rlen);

      move(Ra, Pm_base);  // Save Pm_base in Ra
      restore_regs();  // Restore caller's Pm_base

      // Copy our result into caller's Pm_base
      reverse(Pm_base, Ra, Rlen, t0, t1);

      leave();
      bind(nothing);
      jr(RA);

      return entry;
    }
    // In C, approximately:

    // void
    // montgomery_multiply(unsigned long Pa_base[], unsigned long Pb_base[],
    //                     unsigned long Pn_base[], unsigned long Pm_base[],
    //                     unsigned long inv, int len) {
    //   unsigned long t0 = 0, t1 = 0, t2 = 0; // Triple-precision accumulator
    //   unsigned long Ra, Rb, Rn, Rm;
    //   int i, Iam, Ibn;

    //   assert(inv * Pn_base[0] == -1UL, "broken inverse in Montgomery multiply");

    //   for (i = 0; i < len; i++) {
    //     int j;

    //     Iam = 0;
    //     Ibn = i;

    //     Ra = Pa_base[Iam];
    //     Rb = Pb_base[Iam];
    //     Rm = Pm_base[Ibn];
    //     Rn = Pn_base[Ibn];

    //     int iters = i;
    //     for (j = 0; iters--; j++) {
    //       assert(Ra == Pa_base[j] && Rb == Pb_base[i-j], "must be");
    //       MACC(Ra, Rb, t0, t1, t2);
    //       Ra = Pa_base[++Iam];
    //       Rb = pb_base[--Ibn];
    //       assert(Rm == Pm_base[j] && Rn == Pn_base[i-j], "must be");
    //       MACC(Rm, Rn, t0, t1, t2);
    //       Rm = Pm_base[++Iam];
    //       Rn = Pn_base[--Ibn];
    //     }

    //     assert(Ra == Pa_base[i] && Rb == Pb_base[0], "must be");
    //     MACC(Ra, Rb, t0, t1, t2);
    //     Pm_base[Iam] = Rm = t0 * inv;
    //     assert(Rm == Pm_base[i] && Rn == Pn_base[0], "must be");
    //     MACC(Rm, Rn, t0, t1, t2);

    //     assert(t0 == 0, "broken Montgomery multiply");

    //     t0 = t1; t1 = t2; t2 = 0;
    //   }

    //   for (i = len; i < 2*len; i++) {
    //     int j;

    //     Iam = i - len;
    //     Ibn = len;

    //     Ra = Pa_base[++Iam];
    //     Rb = Pb_base[--Ibn];
    //     Rm = Pm_base[++Iam];
    //     Rn = Pn_base[--Ibn];

    //     int iters = len*2-i-1;
    //     for (j = i-len+1; iters--; j++) {
    //       assert(Ra == Pa_base[j] && Rb == Pb_base[i-j], "must be");
    //       MACC(Ra, Rb, t0, t1, t2);
    //       Ra = Pa_base[++Iam];
    //       Rb = Pb_base[--Ibn];
    //       assert(Rm == Pm_base[j] && Rn == Pn_base[i-j], "must be");
    //       MACC(Rm, Rn, t0, t1, t2);
    //       Rm = Pm_base[++Iam];
    //       Rn = Pn_base[--Ibn];
    //     }

    //     Pm_base[i-len] = t0;
    //     t0 = t1; t1 = t2; t2 = 0;
    //   }

    //   while (t0)
    //     t0 = sub(Pm_base, Pn_base, t0, len);
    // }
  };

  // Initialization
  void generate_initial() {
    // Generates all stubs and initializes the entry points

    //-------------------------------------------------------------
    //-----------------------------------------------------------
    // entry points that exist in all platforms
    // Note: This is code that could be shared among different platforms - however the benefit seems to be smaller
    // than the disadvantage of having a much more complicated generator structure.
    // See also comment in stubRoutines.hpp.
    StubRoutines::_forward_exception_entry = generate_forward_exception();
    StubRoutines::_call_stub_entry = generate_call_stub(StubRoutines::_call_stub_return_address);
    // is referenced by megamorphic call
    StubRoutines::_catch_exception_entry = generate_catch_exception();

    StubRoutines::_handler_for_unsafe_access_entry = generate_handler_for_unsafe_access();

    StubRoutines::_throw_StackOverflowError_entry = generate_throw_exception("StackOverflowError throw_exception",
                                                                              CAST_FROM_FN_PTR(address, SharedRuntime::throw_StackOverflowError),   false);
    // platform dependent
    StubRoutines::la::_get_previous_fp_entry = generate_get_previous_fp();
  }

  void generate_all() {
    // Generates all stubs and initializes the entry points

    // These entry points require SharedInfo::stack0 to be set up in
    // non-core builds and need to be relocatable, so they each
    // fabricate a RuntimeStub internally.
    StubRoutines::_throw_AbstractMethodError_entry = generate_throw_exception("AbstractMethodError throw_exception",
                                                                               CAST_FROM_FN_PTR(address, SharedRuntime::throw_AbstractMethodError),  false);

    StubRoutines::_throw_IncompatibleClassChangeError_entry = generate_throw_exception("IncompatibleClassChangeError throw_exception",
                                                                               CAST_FROM_FN_PTR(address, SharedRuntime:: throw_IncompatibleClassChangeError), false);

    StubRoutines::_throw_NullPointerException_at_call_entry = generate_throw_exception("NullPointerException at call throw_exception",
                                                                                        CAST_FROM_FN_PTR(address, SharedRuntime::throw_NullPointerException_at_call), false);

    // entry points that are platform specific

    // support for verify_oop (must happen after universe_init)
    StubRoutines::_verify_oop_subroutine_entry     = generate_verify_oop();
#ifndef CORE
    // arraycopy stubs used by compilers
    generate_arraycopy_stubs();
#endif

    // Safefetch stubs.
    generate_safefetch("SafeFetch32", sizeof(int),     &StubRoutines::_safefetch32_entry,
                                                       &StubRoutines::_safefetch32_fault_pc,
                                                       &StubRoutines::_safefetch32_continuation_pc);
    generate_safefetch("SafeFetchN", sizeof(intptr_t), &StubRoutines::_safefetchN_entry,
                                                       &StubRoutines::_safefetchN_fault_pc,
                                                       &StubRoutines::_safefetchN_continuation_pc);

    if (UseMontgomeryMultiplyIntrinsic) {
      StubCodeMark mark(this, "StubRoutines", "montgomeryMultiply");
      MontgomeryMultiplyGenerator g(_masm, false /* squaring */);
      StubRoutines::_montgomeryMultiply = g.generate_multiply();
    }

    if (UseMontgomerySquareIntrinsic) {
      StubCodeMark mark(this, "StubRoutines", "montgomerySquare");
      MontgomeryMultiplyGenerator g(_masm, true /* squaring */);
      // We use generate_multiply() rather than generate_square()
      // because it's faster for the sizes of modulus we care about.
      StubRoutines::_montgomerySquare = g.generate_multiply();
    }

  }

 public:
  StubGenerator(CodeBuffer* code, bool all) : StubCodeGenerator(code) {
    if (all) {
      generate_all();
    } else {
      generate_initial();
    }
  }
}; // end class declaration

void StubGenerator_generate(CodeBuffer* code, bool all) {
  StubGenerator g(code, all);
}
